# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_EmptyArray():
    a = ak.contents.emptyarray.EmptyArray()
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 0
    with pytest.raises(IndexError):
        a[0]
    assert isinstance(a[10:20], ak.contents.emptyarray.EmptyArray)
    assert a.to_typetracer()[10:20].form == a[10:20].form
    assert len(a[10:20]) == 0
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]


def test_NumpyArray():
    a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 4
    assert a[2] == 2.2
    assert a[-2] == 2.2
    assert type(a[2]) is np.float64
    assert a.to_typetracer()[2].dtype.type is np.float64
    with pytest.raises(IndexError):
        a[4]
    with pytest.raises(IndexError):
        a[-5]
    assert isinstance(a[2:], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()[2:].form == a[2:].form
    assert a[2:][0] == 2.2
    assert len(a[2:]) == 2
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]

    b = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b) == 2
    assert isinstance(b[1], ak.contents.numpyarray.NumpyArray)
    assert len(b[1]) == 3
    with pytest.raises(IndexError):
        b[2]
    assert b[1][2][0] == 25
    assert len(b[1][2][1:]) == 4
    assert b[1][2][1:][2] == 28
    assert b.to_typetracer()[1][2][1:].form == b[1][2][1:].form
    with pytest.raises(IndexError):
        b["bad"]
    with pytest.raises(IndexError):
        b[["bad", "good", "ok"]]


def test_RegularArray_NumpyArray():
    # 6.6 is inaccessible
    a = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(
            np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
        ),
        3,
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 2
    assert isinstance(a[1], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()[1].form == a[1].form
    assert len(a[1]) == 3
    assert a[1][2] == 5.5
    assert a[-1][2] == 5.5
    assert isinstance(a[1:2], ak.contents.regulararray.RegularArray)
    assert a.to_typetracer()[1:2].form == a[1:2].form
    assert len(a[1:]) == 1
    assert len(a[1:100]) == 1
    with pytest.raises(IndexError):
        a[2]
    with pytest.raises(IndexError):
        a[-3]
    with pytest.raises(IndexError):
        a[1][3]
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]

    b = ak.contents.regulararray.RegularArray(
        ak.contents.emptyarray.EmptyArray(), 0, zeros_length=10
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b) == 10
    assert isinstance(b[5], ak.contents.emptyarray.EmptyArray)
    assert b.to_typetracer()[5].form == b[5].form
    assert len(b[5]) == 0
    assert isinstance(b[7:], ak.contents.regulararray.RegularArray)
    assert b.to_typetracer()[7:].form == b[7:].form
    assert len(b[7:]) == 3
    assert len(b[7:100]) == 3
    with pytest.raises(IndexError):
        b["bad"]
    with pytest.raises(IndexError):
        b[["bad", "good", "ok"]]


def test_ListArray_NumpyArray():
    # 200 is inaccessible in stops
    # 6.6, 7.7, and 8.8 are inaccessible in content
    a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], dtype=np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
        ),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 3
    with pytest.raises(IndexError):
        a[3]
    with pytest.raises(IndexError):
        a[-4]
    assert isinstance(a[2], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()[2].form == a[2].form
    assert len(a[0]) == 3
    assert len(a[1]) == 0
    assert len(a[2]) == 2
    assert len(a[-3]) == 3
    assert len(a[-2]) == 0
    assert len(a[-1]) == 2
    assert a[0][-1] == 3.3
    assert a[2][-1] == 5.5
    assert isinstance(a[1:], ak.contents.listarray.ListArray)
    assert a.to_typetracer()[1:].form == a[1:].form
    assert len(a[1:]) == 2
    assert len(a[-2:]) == 2
    assert len(a[1:100]) == 2
    assert len(a[-2:100]) == 2
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]


def test_ListOffsetArray_NumpyArray():
    # 6.6 and 7.7 are inaccessible
    a = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6])),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
        ),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 3
    with pytest.raises(IndexError):
        a[3]
    with pytest.raises(IndexError):
        a[-4]
    assert isinstance(a[2], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()[2].form == a[2].form
    assert len(a[0]) == 3
    assert len(a[1]) == 0
    assert len(a[2]) == 2
    assert len(a[-3]) == 3
    assert len(a[-2]) == 0
    assert len(a[-1]) == 2
    assert a[0][-1] == 3.3
    assert a[2][-1] == 5.5
    assert isinstance(a[1:], ak.contents.listoffsetarray.ListOffsetArray)
    assert a.to_typetracer()[1:].form == a[1:].form
    assert len(a[1:]) == 2
    assert len(a[-2:]) == 2
    assert len(a[1:100]) == 2
    assert len(a[-2:100]) == 2
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]


def test_RecordArray_NumpyArray():
    # 5.5 is inaccessible
    a = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4])),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 5
    with pytest.raises(IndexError):
        a[5]
    with pytest.raises(IndexError):
        a[-6]
    assert isinstance(a[2], ak.record.Record)
    assert a.to_typetracer()[2].array.form == a[2].array.form
    assert a[2]["y"] == 2.2
    assert a[-3]["y"] == 2.2
    assert isinstance(a[2:], ak.contents.recordarray.RecordArray)
    assert a.to_typetracer()[2:].form == a[2:].form
    assert len(a[2:]) == 3
    assert len(a[-3:]) == 3
    assert len(a[2:100]) == 3
    assert len(a[-3:100]) == 3
    assert isinstance(a["y"], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()["y"].form == a["y"].form
    assert a["y"][2] == 2.2
    assert a["y"][-3] == 2.2
    with pytest.raises(IndexError):
        a["z"]
    with pytest.raises(IndexError):
        a[["x", "z"]]
    assert len(a[["x", "y"]]) == 5
    assert a.to_typetracer()[["x", "y"]].form == a[["x", "y"]].form

    # 5.5 is inaccessible
    b = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4])),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b) == 5
    with pytest.raises(IndexError):
        b[5]
    with pytest.raises(IndexError):
        b[-6]
    assert isinstance(b[2], ak.record.Record)
    assert b.to_typetracer()[2].array.form == b[2].array.form
    assert b[2]["1"] == 2.2
    assert b[-3]["1"] == 2.2
    assert isinstance(b[2:], ak.contents.recordarray.RecordArray)
    assert b.to_typetracer()[2:].form == b[2:].form
    assert len(b[2:]) == 3
    assert len(b[-3:]) == 3
    assert len(b[2:100]) == 3
    assert len(b[-3:100]) == 3
    assert isinstance(b["1"], ak.contents.numpyarray.NumpyArray)
    assert b.to_typetracer()["1"].form == b["1"].form
    assert b["1"][2] == 2.2
    assert b["1"][-3] == 2.2
    with pytest.raises(IndexError):
        a["2"]
    assert len(b[["0", "1"]]) == 5
    assert b.to_typetracer()[["0", "1"]].form == b[["0", "1"]].form

    c = ak.contents.recordarray.RecordArray([], [], 10)
    assert c.to_typetracer().form == c.form
    assert c.to_typetracer().form.type == c.form.type
    assert len(c) == 10
    assert isinstance(c[5], ak.record.Record)
    assert c.to_typetracer()[5].array.form == c[5].array.form
    assert isinstance(c[7:], ak.contents.recordarray.RecordArray)
    assert c.to_typetracer()[7:].form == c[7:].form
    assert len(c[7:]) == 3
    assert len(c[-3:]) == 3
    with pytest.raises(IndexError):
        c["x"]

    d = ak.contents.recordarray.RecordArray([], None, 10)
    assert d.to_typetracer().form == d.form
    assert d.to_typetracer().form.type == d.form.type
    assert len(d) == 10
    assert isinstance(d[5], ak.record.Record)
    assert d.to_typetracer()[5].array.form == d[5].array.form
    assert isinstance(d[7:], ak.contents.recordarray.RecordArray)
    assert d.to_typetracer()[7:].form == d[7:].form
    assert len(d[7:]) == 3
    assert len(d[-3:]) == 3
    with pytest.raises(IndexError):
        d["0"]


def test_IndexedArray_NumpyArray():
    # 4.4 is inaccessible; 3.3 and 5.5 appear twice
    a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4])),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 7
    assert a[0] == 3.3
    assert a[1] == 3.3
    assert a[2] == 1.1
    assert a[3] == 2.2
    assert a[4] == 5.5
    assert a[5] == 6.6
    assert a[6] == 5.5
    assert a[-7] == 3.3
    assert a[-6] == 3.3
    assert a[-5] == 1.1
    assert a[-4] == 2.2
    assert a[-3] == 5.5
    assert a[-2] == 6.6
    assert a[-1] == 5.5
    with pytest.raises(IndexError):
        a[7]
    with pytest.raises(IndexError):
        a[-8]
    assert isinstance(a[3:], ak.contents.indexedarray.IndexedArray)
    assert a.to_typetracer()[3:].form == a[3:].form
    assert len(a[3:]) == 4
    assert len(a[-4:]) == 4
    assert len(a[3:100]) == 4
    assert len(a[-4:100]) == 4
    assert a[3:][1] == 5.5
    assert a[-4:][1] == 5.5
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]


def test_IndexedOptionArray_NumpyArray():
    # 1.1 and 4.4 are inaccessible; 3.3 appears twice
    a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4])),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 7
    assert a[0] == 3.3
    assert a[1] == 3.3
    assert a[2] is None
    assert a[3] == 2.2
    assert a[4] is None
    assert a[5] == 6.6
    assert a[6] == 5.5
    assert a[-7] == 3.3
    assert a[-6] == 3.3
    assert a[-5] is None
    assert a[-4] == 2.2
    assert a[-3] is None
    assert a[-2] == 6.6
    assert a[-1] == 5.5
    with pytest.raises(IndexError):
        a[7]
    with pytest.raises(IndexError):
        a[-8]
    assert isinstance(a[3:], ak.contents.indexedoptionarray.IndexedOptionArray)
    assert a.to_typetracer()[3:].form == a[3:].form
    assert len(a[3:]) == 4
    assert len(a[-4:]) == 4
    assert len(a[3:100]) == 4
    assert len(a[-4:100]) == 4
    assert a[3:][1] is None
    assert a[-4:][1] is None
    assert a[3:][2] == 6.6
    assert a[-4:][2] == 6.6
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]


def test_ByteMaskedArray_NumpyArray():
    # 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], dtype=np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=True,
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 5
    with pytest.raises(IndexError):
        a[5]
    with pytest.raises(IndexError):
        a[-6]
    assert a[0] == 1.1
    assert a[1] is None
    assert a[2] == 3.3
    assert a[3] is None
    assert a[4] == 5.5
    assert a[-5] == 1.1
    assert a[-4] is None
    assert a[-3] == 3.3
    assert a[-2] is None
    assert a[-1] == 5.5
    assert isinstance(a[2:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert a.to_typetracer()[2:].form == a[2:].form
    assert len(a[2:]) == 3
    assert len(a[-3:]) == 3
    assert len(a[2:100]) == 3
    assert len(a[-3:100]) == 3
    assert a[2:][1] is None
    assert a[-3:][1] is None
    assert a[2:][2] == 5.5
    assert a[-3:][2] == 5.5
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]

    # 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], dtype=np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=False,
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b) == 5
    with pytest.raises(IndexError):
        b[5]
    with pytest.raises(IndexError):
        b[-6]
    assert b[0] == 1.1
    assert b[1] is None
    assert b[2] == 3.3
    assert b[3] is None
    assert b[4] == 5.5
    assert b[-5] == 1.1
    assert b[-4] is None
    assert b[-3] == 3.3
    assert b[-2] is None
    assert b[-1] == 5.5
    assert isinstance(b[2:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert b.to_typetracer()[2:].form == b[2:].form
    assert len(b[2:]) == 3
    assert len(b[-3:]) == 3
    assert len(b[2:100]) == 3
    assert len(b[-3:100]) == 3
    assert b[2:][1] is None
    assert b[-3:][1] is None
    assert b[2:][2] == 5.5
    assert b[-3:][2] == 5.5
    with pytest.raises(IndexError):
        b["bad"]
    with pytest.raises(IndexError):
        b[["bad", "good", "ok"]]


def test_BitMaskedArray_NumpyArray():
    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 13
    with pytest.raises(IndexError):
        a[13]
    with pytest.raises(IndexError):
        a[-14]
    assert a[0] == 0.0
    assert a[1] == 1.0
    assert a[2] == 2.0
    assert a[3] == 3.0
    assert a[4] is None
    assert a[5] is None
    assert a[6] is None
    assert a[7] is None
    assert a[8] == 1.1
    assert a[9] is None
    assert a[10] == 3.3
    assert a[11] is None
    assert a[12] == 5.5
    assert a[-13] == 0.0
    assert a[-12] == 1.0
    assert a[-11] == 2.0
    assert a[-10] == 3.0
    assert a[-9] is None
    assert a[-8] is None
    assert a[-7] is None
    assert a[-6] is None
    assert a[-5] == 1.1
    assert a[-4] is None
    assert a[-3] == 3.3
    assert a[-2] is None
    assert a[-1] == 5.5
    assert isinstance(a[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert a.to_typetracer()[5:].form == a[5:].form
    assert len(a[5:]) == 8
    assert len(a[-8:]) == 8
    assert len(a[5:100]) == 8
    assert len(a[-8:100]) == 8
    assert a[5:][2] is None
    assert a[5:][3] == 1.1
    assert a[-8:][2] is None
    assert a[-8:][3] == 1.1
    with pytest.raises(IndexError):
        a["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b) == 13
    with pytest.raises(IndexError):
        b[13]
    with pytest.raises(IndexError):
        b[-14]
    assert b[0] == 0.0
    assert b[1] == 1.0
    assert b[2] == 2.0
    assert b[3] == 3.0
    assert b[4] is None
    assert b[5] is None
    assert b[6] is None
    assert b[7] is None
    assert b[8] == 1.1
    assert b[9] is None
    assert b[10] == 3.3
    assert b[11] is None
    assert b[12] == 5.5
    assert b[-13] == 0.0
    assert b[-12] == 1.0
    assert b[-11] == 2.0
    assert b[-10] == 3.0
    assert b[-9] is None
    assert b[-8] is None
    assert b[-7] is None
    assert b[-6] is None
    assert b[-5] == 1.1
    assert b[-4] is None
    assert b[-3] == 3.3
    assert b[-2] is None
    assert b[-1] == 5.5
    assert isinstance(b[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert b.to_typetracer()[5:].form == b[5:].form
    assert len(b[5:]) == 8
    assert len(b[-8:]) == 8
    assert len(b[5:100]) == 8
    assert len(b[-8:100]) == 8
    assert b[5:][2] is None
    assert b[5:][3] == 1.1
    assert b[-8:][2] is None
    assert b[-8:][3] == 1.1
    with pytest.raises(IndexError):
        b["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    assert c.to_typetracer().form == c.form
    assert c.to_typetracer().form.type == c.form.type
    assert len(c) == 13
    with pytest.raises(IndexError):
        c[13]
    with pytest.raises(IndexError):
        c[-14]
    assert c[0] == 0.0
    assert c[1] == 1.0
    assert c[2] == 2.0
    assert c[3] == 3.0
    assert c[4] is None
    assert c[5] is None
    assert c[6] is None
    assert c[7] is None
    assert c[8] == 1.1
    assert c[9] is None
    assert c[10] == 3.3
    assert c[11] is None
    assert c[12] == 5.5
    assert c[-13] == 0.0
    assert c[-12] == 1.0
    assert c[-11] == 2.0
    assert c[-10] == 3.0
    assert c[-9] is None
    assert c[-8] is None
    assert c[-7] is None
    assert c[-6] is None
    assert c[-5] == 1.1
    assert c[-4] is None
    assert c[-3] == 3.3
    assert c[-2] is None
    assert c[-1] == 5.5
    assert isinstance(c[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert c.to_typetracer()[5:].form == c[5:].form
    assert len(c[5:]) == 8
    assert len(c[-8:]) == 8
    assert len(c[5:100]) == 8
    assert len(c[-8:100]) == 8
    assert c[5:][2] is None
    assert c[5:][3] == 1.1
    assert c[-8:][2] is None
    assert c[-8:][3] == 1.1
    with pytest.raises(IndexError):
        c["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    assert d.to_typetracer().form == d.form
    assert d.to_typetracer().form.type == d.form.type
    assert len(d) == 13
    with pytest.raises(IndexError):
        d[13]
    with pytest.raises(IndexError):
        d[-14]
    assert d[0] == 0.0
    assert d[1] == 1.0
    assert d[2] == 2.0
    assert d[3] == 3.0
    assert d[4] is None
    assert d[5] is None
    assert d[6] is None
    assert d[7] is None
    assert d[8] == 1.1
    assert d[9] is None
    assert d[10] == 3.3
    assert d[11] is None
    assert d[12] == 5.5
    assert d[-13] == 0.0
    assert d[-12] == 1.0
    assert d[-11] == 2.0
    assert d[-10] == 3.0
    assert d[-9] is None
    assert d[-8] is None
    assert d[-7] is None
    assert d[-6] is None
    assert d[-5] == 1.1
    assert d[-4] is None
    assert d[-3] == 3.3
    assert d[-2] is None
    assert d[-1] == 5.5
    assert isinstance(d[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert d.to_typetracer()[5:].form == d[5:].form
    assert len(d[5:]) == 8
    assert len(d[-8:]) == 8
    assert len(d[5:100]) == 8
    assert len(d[-8:100]) == 8
    assert d[5:][2] is None
    assert d[5:][3] == 1.1
    assert d[-8:][2] is None
    assert d[-8:][3] == 1.1
    with pytest.raises(IndexError):
        d["bad"]


def test_UnmaskedArray_NumpyArray():
    a = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.numpyarray.NumpyArray(
            np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
        )
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 4
    assert a[2] == 2.2
    assert a[-2] == 2.2
    assert type(a[2]) is np.float64
    with pytest.raises(IndexError):
        a[4]
    with pytest.raises(IndexError):
        a[-5]
    assert isinstance(a[2:], ak.contents.unmaskedarray.UnmaskedArray)
    assert a[2:][0] == 2.2
    assert len(a[2:]) == 2
    with pytest.raises(IndexError):
        a["bad"]


def test_UnionArray_NumpyArray():
    # 100 is inaccessible in index
    # 1.1 is inaccessible in contents[1]
    a = ak.contents.unionarray.UnionArray.simplified(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3])),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a) == 7
    with pytest.raises(IndexError):
        a[7]
    with pytest.raises(IndexError):
        a[-8]
    assert a[0] == 5.5
    assert a[1] == 4.4
    assert a[2] == 1.0
    assert a[3] == 2.0
    assert a[4] == 3.3
    assert a[5] == 3.0
    assert a[6] == 5.5
    assert a[-7] == 5.5
    assert a[-6] == 4.4
    assert a[-5] == 1.0
    assert a[-4] == 2.0
    assert a[-3] == 3.3
    assert a[-2] == 3.0
    assert a[-1] == 5.5
    assert a.to_typetracer()[3:].form == a[3:].form
    assert len(a[3:]) == 4
    assert len(a[-4:]) == 4
    assert len(a[3:100]) == 4
    assert len(a[-4:100]) == 4
    assert a[3:][1] == 3.3
    assert a[3:][2] == 3.0
    assert a[-4:][1] == 3.3
    assert a[-4:][2] == 3.0
    with pytest.raises(IndexError):
        a["bad"]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"]]


def test_RegularArray_RecordArray_NumpyArray():
    # 6.6 is inaccessible
    a = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        3,
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 2
    assert a.to_typetracer()["nest"].form == a["nest"].form
    assert isinstance(a["nest"][1], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()["nest"][1].form == a["nest"][1].form
    assert len(a["nest"][1]) == 3
    assert a["nest"][1][2] == 5.5
    assert a["nest"][-1][2] == 5.5
    assert isinstance(a["nest"][1:2], ak.contents.regulararray.RegularArray)
    assert a.to_typetracer()["nest"][1:2].form == a["nest"][1:2].form
    assert len(a["nest"][1:]) == 1
    assert len(a["nest"][1:100]) == 1
    with pytest.raises(IndexError):
        a["nest"][2]
    with pytest.raises(IndexError):
        a["nest"][-3]
    with pytest.raises(IndexError):
        a["nest"][1][3]
    with pytest.raises(IndexError):
        a["nest"]["bad"]

    b = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [ak.contents.emptyarray.EmptyArray()], ["nest"]
        ),
        0,
        zeros_length=10,
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b["nest"]) == 10
    assert b.to_typetracer()["nest"].form == b["nest"].form
    assert isinstance(b["nest"][5], ak.contents.emptyarray.EmptyArray)
    assert b.to_typetracer()["nest"][5].form == b["nest"][5].form
    assert len(b["nest"][5]) == 0
    assert isinstance(b["nest"][7:], ak.contents.regulararray.RegularArray)
    assert b.to_typetracer()["nest"][7:].form == b["nest"][7:].form
    assert len(b["nest"][7:]) == 3
    assert len(b["nest"][7:100]) == 3
    with pytest.raises(IndexError):
        b["nest"]["bad"]


def test_ListArray_RecordArray_NumpyArray():
    # 200 is inaccessible in stops
    # 6.6, 7.7, and 8.8 are inaccessible in content
    a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1])),
        ak.index.Index(np.array([7, 100, 3, 200])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
                )
            ],
            ["nest"],
        ),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 3
    assert a.to_typetracer()["nest"].form == a["nest"].form
    with pytest.raises(IndexError):
        a["nest"][3]
    with pytest.raises(IndexError):
        a["nest"][-4]
    assert isinstance(a["nest"][2], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()["nest"][2].form == a["nest"][2].form
    assert len(a["nest"][0]) == 3
    assert len(a["nest"][1]) == 0
    assert len(a["nest"][2]) == 2
    assert len(a["nest"][-3]) == 3
    assert len(a["nest"][-2]) == 0
    assert len(a["nest"][-1]) == 2
    assert a["nest"][0][-1] == 3.3
    assert a["nest"][2][-1] == 5.5
    assert isinstance(a["nest"][1:], ak.contents.listarray.ListArray)
    assert a.to_typetracer()["nest"][1:].form == a["nest"][1:].form
    assert len(a["nest"][1:]) == 2
    assert len(a["nest"][-2:]) == 2
    assert len(a["nest"][1:100]) == 2
    assert len(a["nest"][-2:100]) == 2
    with pytest.raises(IndexError):
        a["nest"]["bad"]


def test_ListOffsetArray_RecordArray_NumpyArray():
    # 6.6 and 7.7 are inaccessible
    a = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
                )
            ],
            ["nest"],
        ),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 3
    assert a.to_typetracer()["nest"].form == a["nest"].form
    with pytest.raises(IndexError):
        a["nest"][3]
    with pytest.raises(IndexError):
        a["nest"][-4]
    assert isinstance(a["nest"][2], ak.contents.numpyarray.NumpyArray)
    assert a.to_typetracer()["nest"][2].form == a["nest"][2].form
    assert len(a["nest"][0]) == 3
    assert len(a["nest"][1]) == 0
    assert len(a["nest"][2]) == 2
    assert len(a["nest"][-3]) == 3
    assert len(a["nest"][-2]) == 0
    assert len(a["nest"][-1]) == 2
    assert a["nest"][0][-1] == 3.3
    assert a["nest"][2][-1] == 5.5
    assert isinstance(a["nest"][1:], ak.contents.listoffsetarray.ListOffsetArray)
    assert a.to_typetracer()["nest"][1:].form == a["nest"][1:].form
    assert len(a["nest"][1:]) == 2
    assert len(a["nest"][-2:]) == 2
    assert len(a["nest"][1:100]) == 2
    assert len(a["nest"][-2:100]) == 2
    with pytest.raises(IndexError):
        a["nest"]["bad"]


def test_IndexedArray_RecordArray_NumpyArray():
    # 4.4 is inaccessible; 3.3 and 5.5 appear twice
    a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 7
    assert a.to_typetracer()["nest"].form == a["nest"].form
    assert a["nest"][0] == 3.3
    assert a["nest"][1] == 3.3
    assert a["nest"][2] == 1.1
    assert a["nest"][3] == 2.2
    assert a["nest"][4] == 5.5
    assert a["nest"][5] == 6.6
    assert a["nest"][6] == 5.5
    assert a["nest"][-7] == 3.3
    assert a["nest"][-6] == 3.3
    assert a["nest"][-5] == 1.1
    assert a["nest"][-4] == 2.2
    assert a["nest"][-3] == 5.5
    assert a["nest"][-2] == 6.6
    assert a["nest"][-1] == 5.5
    with pytest.raises(IndexError):
        a["nest"][7]
    with pytest.raises(IndexError):
        a["nest"][-8]
    assert isinstance(a["nest"][3:], ak.contents.indexedarray.IndexedArray)
    assert a.to_typetracer()["nest"][3:].form == a["nest"][3:].form
    assert len(a["nest"][3:]) == 4
    assert len(a["nest"][-4:]) == 4
    assert len(a["nest"][3:100]) == 4
    assert len(a["nest"][-4:100]) == 4
    assert a["nest"][3:][1] == 5.5
    assert a["nest"][-4:][1] == 5.5
    with pytest.raises(IndexError):
        a["nest"]["bad"]


def test_IndexedOptionArray_RecordArray_NumpyArray():
    # 1.1 and 4.4 are inaccessible; 3.3 appears twice
    a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 7
    assert a.to_typetracer()["nest"].form == a["nest"].form
    assert a["nest"][0] == 3.3
    assert a["nest"][1] == 3.3
    assert a["nest"][2] is None
    assert a["nest"][3] == 2.2
    assert a["nest"][4] is None
    assert a["nest"][5] == 6.6
    assert a["nest"][6] == 5.5
    assert a["nest"][-7] == 3.3
    assert a["nest"][-6] == 3.3
    assert a["nest"][-5] is None
    assert a["nest"][-4] == 2.2
    assert a["nest"][-3] is None
    assert a["nest"][-2] == 6.6
    assert a["nest"][-1] == 5.5
    with pytest.raises(IndexError):
        a["nest"][7]
    with pytest.raises(IndexError):
        a["nest"][-8]
    assert isinstance(a["nest"][3:], ak.contents.indexedoptionarray.IndexedOptionArray)
    assert a.to_typetracer()["nest"][3:].form == a["nest"][3:].form
    assert len(a["nest"][3:]) == 4
    assert len(a["nest"][-4:]) == 4
    assert len(a["nest"][3:100]) == 4
    assert len(a["nest"][-4:100]) == 4
    assert a["nest"][3:][1] is None
    assert a["nest"][-4:][1] is None
    assert a["nest"][3:][2] == 6.6
    assert a["nest"][-4:][2] == 6.6
    with pytest.raises(IndexError):
        a["nest"]["bad"]


def test_ByteMaskedArray_RecordArray_NumpyArray():
    # 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], dtype=np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=True,
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 5
    assert a.to_typetracer()["nest"].form == a["nest"].form
    with pytest.raises(IndexError):
        a["nest"][5]
    with pytest.raises(IndexError):
        a["nest"][-6]
    assert a["nest"][0] == 1.1
    assert a["nest"][1] is None
    assert a["nest"][2] == 3.3
    assert a["nest"][3] is None
    assert a["nest"][4] == 5.5
    assert a["nest"][-5] == 1.1
    assert a["nest"][-4] is None
    assert a["nest"][-3] == 3.3
    assert a["nest"][-2] is None
    assert a["nest"][-1] == 5.5
    assert isinstance(a["nest"][2:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert a.to_typetracer()["nest"][2:].form == a["nest"][2:].form
    assert len(a["nest"][2:]) == 3
    assert len(a["nest"][-3:]) == 3
    assert len(a["nest"][2:100]) == 3
    assert len(a["nest"][-3:100]) == 3
    assert a["nest"][2:][1] is None
    assert a["nest"][-3:][1] is None
    assert a["nest"][2:][2] == 5.5
    assert a["nest"][-3:][2] == 5.5
    with pytest.raises(IndexError):
        a["nest"]["bad"]

    # 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], dtype=np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=False,
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b["nest"]) == 5
    assert b.to_typetracer()["nest"].form == b["nest"].form
    with pytest.raises(IndexError):
        b["nest"][5]
    with pytest.raises(IndexError):
        b["nest"][-6]
    assert b["nest"][0] == 1.1
    assert b["nest"][1] is None
    assert b["nest"][2] == 3.3
    assert b["nest"][3] is None
    assert b["nest"][4] == 5.5
    assert b["nest"][-5] == 1.1
    assert b["nest"][-4] is None
    assert b["nest"][-3] == 3.3
    assert b["nest"][-2] is None
    assert b["nest"][-1] == 5.5
    assert isinstance(b["nest"][2:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert b.to_typetracer()["nest"][2:].form == b["nest"][2:].form
    assert len(b["nest"][2:]) == 3
    assert len(b["nest"][-3:]) == 3
    assert len(b["nest"][2:100]) == 3
    assert len(b["nest"][-3:100]) == 3
    assert b["nest"][2:][1] is None
    assert b["nest"][-3:][1] is None
    assert b["nest"][2:][2] == 5.5
    assert b["nest"][-3:][2] == 5.5
    with pytest.raises(IndexError):
        b["nest"]["bad"]


def test_BitMaskedArray_RecordArray_NumpyArray():
    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        True,
                        True,
                        True,
                        True,
                        False,
                        False,
                        False,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                    ]
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 13
    assert a.to_typetracer()["nest"].form == a["nest"].form
    with pytest.raises(IndexError):
        a["nest"][13]
    with pytest.raises(IndexError):
        a["nest"][-14]
    assert a["nest"][0] == 0.0
    assert a["nest"][1] == 1.0
    assert a["nest"][2] == 2.0
    assert a["nest"][3] == 3.0
    assert a["nest"][4] is None
    assert a["nest"][5] is None
    assert a["nest"][6] is None
    assert a["nest"][7] is None
    assert a["nest"][8] == 1.1
    assert a["nest"][9] is None
    assert a["nest"][10] == 3.3
    assert a["nest"][11] is None
    assert a["nest"][12] == 5.5
    assert a["nest"][-13] == 0.0
    assert a["nest"][-12] == 1.0
    assert a["nest"][-11] == 2.0
    assert a["nest"][-10] == 3.0
    assert a["nest"][-9] is None
    assert a["nest"][-8] is None
    assert a["nest"][-7] is None
    assert a["nest"][-6] is None
    assert a["nest"][-5] == 1.1
    assert a["nest"][-4] is None
    assert a["nest"][-3] == 3.3
    assert a["nest"][-2] is None
    assert a["nest"][-1] == 5.5
    assert isinstance(a["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert a.to_typetracer()["nest"][5:].form == a["nest"][5:].form
    assert len(a["nest"][5:]) == 8
    assert len(a["nest"][-8:]) == 8
    assert len(a["nest"][5:100]) == 8
    assert len(a["nest"][-8:100]) == 8
    assert a["nest"][5:][2] is None
    assert a["nest"][5:][3] == 1.1
    assert a["nest"][-8:][2] is None
    assert a["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        a["nest"]["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    assert b.to_typetracer().form == b.form
    assert b.to_typetracer().form.type == b.form.type
    assert len(b["nest"]) == 13
    assert b.to_typetracer()["nest"].form == b["nest"].form
    with pytest.raises(IndexError):
        b["nest"][13]
    with pytest.raises(IndexError):
        b["nest"][-14]
    assert b["nest"][0] == 0.0
    assert b["nest"][1] == 1.0
    assert b["nest"][2] == 2.0
    assert b["nest"][3] == 3.0
    assert b["nest"][4] is None
    assert b["nest"][5] is None
    assert b["nest"][6] is None
    assert b["nest"][7] is None
    assert b["nest"][8] == 1.1
    assert b["nest"][9] is None
    assert b["nest"][10] == 3.3
    assert b["nest"][11] is None
    assert b["nest"][12] == 5.5
    assert b["nest"][-13] == 0.0
    assert b["nest"][-12] == 1.0
    assert b["nest"][-11] == 2.0
    assert b["nest"][-10] == 3.0
    assert b["nest"][-9] is None
    assert b["nest"][-8] is None
    assert b["nest"][-7] is None
    assert b["nest"][-6] is None
    assert b["nest"][-5] == 1.1
    assert b["nest"][-4] is None
    assert b["nest"][-3] == 3.3
    assert b["nest"][-2] is None
    assert b["nest"][-1] == 5.5
    assert isinstance(b["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert b.to_typetracer()["nest"][5:].form == b["nest"][5:].form
    assert len(b["nest"][5:]) == 8
    assert len(b["nest"][-8:]) == 8
    assert len(b["nest"][5:100]) == 8
    assert len(b["nest"][-8:100]) == 8
    assert b["nest"][5:][2] is None
    assert b["nest"][5:][3] == 1.1
    assert b["nest"][-8:][2] is None
    assert b["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        b["nest"]["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    assert c.to_typetracer().form == c.form
    assert c.to_typetracer().form.type == c.form.type
    assert len(c["nest"]) == 13
    assert c.to_typetracer()["nest"].form == c["nest"].form
    with pytest.raises(IndexError):
        c["nest"][13]
    with pytest.raises(IndexError):
        c["nest"][-14]
    assert c["nest"][0] == 0.0
    assert c["nest"][1] == 1.0
    assert c["nest"][2] == 2.0
    assert c["nest"][3] == 3.0
    assert c["nest"][4] is None
    assert c["nest"][5] is None
    assert c["nest"][6] is None
    assert c["nest"][7] is None
    assert c["nest"][8] == 1.1
    assert c["nest"][9] is None
    assert c["nest"][10] == 3.3
    assert c["nest"][11] is None
    assert c["nest"][12] == 5.5
    assert c["nest"][-13] == 0.0
    assert c["nest"][-12] == 1.0
    assert c["nest"][-11] == 2.0
    assert c["nest"][-10] == 3.0
    assert c["nest"][-9] is None
    assert c["nest"][-8] is None
    assert c["nest"][-7] is None
    assert c["nest"][-6] is None
    assert c["nest"][-5] == 1.1
    assert c["nest"][-4] is None
    assert c["nest"][-3] == 3.3
    assert c["nest"][-2] is None
    assert c["nest"][-1] == 5.5
    assert isinstance(c["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert c.to_typetracer()["nest"][5:].form == c["nest"][5:].form
    assert len(c["nest"][5:]) == 8
    assert len(c["nest"][-8:]) == 8
    assert len(c["nest"][5:100]) == 8
    assert len(c["nest"][-8:100]) == 8
    assert c["nest"][5:][2] is None
    assert c["nest"][5:][3] == 1.1
    assert c["nest"][-8:][2] is None
    assert c["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        c["nest"]["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    assert d.to_typetracer().form == d.form
    assert d.to_typetracer().form.type == d.form.type
    assert len(d["nest"]) == 13
    assert d.to_typetracer()["nest"].form == d["nest"].form
    with pytest.raises(IndexError):
        d["nest"][13]
    with pytest.raises(IndexError):
        d["nest"][-14]
    assert d["nest"][0] == 0.0
    assert d["nest"][1] == 1.0
    assert d["nest"][2] == 2.0
    assert d["nest"][3] == 3.0
    assert d["nest"][4] is None
    assert d["nest"][5] is None
    assert d["nest"][6] is None
    assert d["nest"][7] is None
    assert d["nest"][8] == 1.1
    assert d["nest"][9] is None
    assert d["nest"][10] == 3.3
    assert d["nest"][11] is None
    assert d["nest"][12] == 5.5
    assert d["nest"][-13] == 0.0
    assert d["nest"][-12] == 1.0
    assert d["nest"][-11] == 2.0
    assert d["nest"][-10] == 3.0
    assert d["nest"][-9] is None
    assert d["nest"][-8] is None
    assert d["nest"][-7] is None
    assert d["nest"][-6] is None
    assert d["nest"][-5] == 1.1
    assert d["nest"][-4] is None
    assert d["nest"][-3] == 3.3
    assert d["nest"][-2] is None
    assert d["nest"][-1] == 5.5
    assert isinstance(d["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert d.to_typetracer()["nest"][5:].form == d["nest"][5:].form
    assert len(d["nest"][5:]) == 8
    assert len(d["nest"][-8:]) == 8
    assert len(d["nest"][5:100]) == 8
    assert len(d["nest"][-8:100]) == 8
    assert d["nest"][5:][2] is None
    assert d["nest"][5:][3] == 1.1
    assert d["nest"][-8:][2] is None
    assert d["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        d["nest"]["bad"]


def test_UnmaskedArray_RecordArray_NumpyArray():
    a = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
                )
            ],
            ["nest"],
        )
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 4
    assert a.to_typetracer()["nest"].form == a["nest"].form
    assert a["nest"][2] == 2.2
    assert a["nest"][-2] == 2.2
    assert type(a["nest"][2]) is np.float64
    with pytest.raises(IndexError):
        a["nest"][4]
    with pytest.raises(IndexError):
        a["nest"][-5]
    assert isinstance(a["nest"][2:], ak.contents.unmaskedarray.UnmaskedArray)
    assert a.to_typetracer()["nest"][2:].form == a["nest"][2:].form
    assert a["nest"][2:][0] == 2.2
    assert len(a["nest"][2:]) == 2
    with pytest.raises(IndexError):
        a["nest"]["bad"]


def test_UnionArray_RecordArray_NumpyArray():
    # 100 is inaccessible in index
    # 1.1 is inaccessible in contents[1]
    a = ak.contents.unionarray.UnionArray.simplified(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.recordarray.RecordArray(
                [ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3]))], ["nest"]
            ),
            ak.contents.recordarray.RecordArray(
                [
                    ak.contents.numpyarray.NumpyArray(
                        np.array([1.1, 2.2, 3.3, 4.4, 5.5])
                    )
                ],
                ["nest"],
            ),
        ],
    )
    assert a.to_typetracer().form == a.form
    assert a.to_typetracer().form.type == a.form.type
    assert len(a["nest"]) == 7
    assert a.to_typetracer()["nest"].form == a["nest"].form
    with pytest.raises(IndexError):
        a["nest"][7]
    with pytest.raises(IndexError):
        a["nest"][-8]
    assert a["nest"][0] == 5.5
    assert a["nest"][1] == 4.4
    assert a["nest"][2] == 1.0
    assert a["nest"][3] == 2.0
    assert a["nest"][4] == 3.3
    assert a["nest"][5] == 3.0
    assert a["nest"][6] == 5.5
    assert a["nest"][-7] == 5.5
    assert a["nest"][-6] == 4.4
    assert a["nest"][-5] == 1.0
    assert a["nest"][-4] == 2.0
    assert a["nest"][-3] == 3.3
    assert a["nest"][-2] == 3.0
    assert a["nest"][-1] == 5.5
    assert a.to_typetracer()["nest"][3:].form == a["nest"][3:].form
    assert len(a["nest"][3:]) == 4
    assert len(a["nest"][-4:]) == 4
    assert len(a["nest"][3:100]) == 4
    assert len(a["nest"][-4:100]) == 4
    assert a["nest"][3:][1] == 3.3
    assert a["nest"][3:][2] == 3.0
    assert a["nest"][-4:][1] == 3.3
    assert a["nest"][-4:][2] == 3.0
    with pytest.raises(IndexError):
        a["nest"]["bad"]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from collections.abc import Iterable

import awkward as ak
from awkward._errors import deprecate
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._parameters import type_parameters_equal
from awkward._typing import JSONSerializable, Self, final
from awkward._util import UNSET
from awkward.forms.form import Form

np = NumpyMetadata.instance()


def from_dtype(dtype, parameters=None, *, time_units_as_parameter: bool = UNSET):
    if dtype.subdtype is None:
        inner_shape = ()
    else:
        inner_shape = dtype.shape
        dtype = dtype.subdtype[0]

    if time_units_as_parameter is UNSET:
        time_units_as_parameter = True

    if time_units_as_parameter:
        deprecate(
            "from_dtype conversion of temporal units to generic `datetime64` and `timedelta64` types is deprecated, "
            "pass `time_units_as_parameter=False` to disable this warning.",
            version="2.4.0",
        )

    if time_units_as_parameter and issubclass(
        dtype.type, (np.datetime64, np.timedelta64)
    ):
        unit, step = np.datetime_data(dtype)
        if unit != "generic":
            unitstr = ("" if step == 1 else str(step)) + unit
            if parameters is None:
                parameters = {}
            else:
                parameters = parameters.copy()
            parameters["__unit__"] = unitstr
            dtype = np.dtype(dtype.type)

    return NumpyForm(
        primitive=ak.types.numpytype.dtype_to_primitive(dtype),
        parameters=parameters,
        inner_shape=inner_shape,
    )


@final
class NumpyForm(Form):
    is_numpy = True

    def __init__(
        self,
        primitive,
        inner_shape=(),
        *,
        parameters=None,
        form_key=None,
    ):
        primitive = ak.types.numpytype.dtype_to_primitive(
            ak.types.numpytype.primitive_to_dtype(primitive)
        )
        if not isinstance(inner_shape, Iterable):
            raise TypeError(
                "{} 'inner_shape' must be iterable, not {}".format(
                    type(self).__name__, repr(inner_shape)
                )
            )

        self._primitive = primitive
        self._inner_shape = tuple(inner_shape)
        self._init(parameters=parameters, form_key=form_key)

    @property
    def primitive(self):
        return self._primitive

    @property
    def inner_shape(self):
        return self._inner_shape

    def copy(
        self,
        primitive=UNSET,
        inner_shape=UNSET,
        *,
        parameters=UNSET,
        form_key=UNSET,
    ):
        return NumpyForm(
            self._primitive if primitive is UNSET else primitive,
            self._inner_shape if inner_shape is UNSET else inner_shape,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(
        cls,
        primitive,
        inner_shape=(),
        *,
        parameters=None,
        form_key=None,
    ):
        return cls(primitive, inner_shape, parameters=parameters, form_key=form_key)

    @property
    def itemsize(self):
        return ak.types.numpytype.primitive_to_dtype(self._primitive).itemsize

    def __repr__(self):
        args = [repr(self._primitive)]
        if len(self._inner_shape) > 0:
            args.append("inner_shape=" + repr(self._inner_shape))
        args += self._repr_args()
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _to_dict_part(self, verbose, toplevel):
        if (
            not verbose
            and not toplevel
            and len(self._inner_shape) == 0
            and (self._parameters is None or len(self._parameters) == 0)
            and self._form_key is None
        ):
            return self._primitive

        else:
            out = {
                "class": "NumpyArray",
                "primitive": self._primitive,
            }
            if verbose or len(self._inner_shape) > 0:
                out["inner_shape"] = list(self._inner_shape)
            return self._to_dict_extra(out, verbose)

    @property
    def type(self):
        out = ak.types.NumpyType(
            self._primitive,
            parameters=None,
        )
        for x in self._inner_shape[::-1]:
            out = ak.types.RegularType(out, x)

        out._parameters = self._parameters

        return out

    def __eq__(self, other):
        if isinstance(other, NumpyForm):
            return (
                self._form_key == other._form_key
                and self._primitive == other._primitive
                and self._inner_shape == other._inner_shape
                and type_parameters_equal(self._parameters, other._parameters)
            )
        else:
            return False

    def to_RegularForm(self):
        out = NumpyForm(self._primitive, (), parameters=None, form_key=None)
        for x in self._inner_shape[::-1]:
            out = ak.forms.RegularForm(out, x, parameters=None, form_key=None)
        out._parameters = self._parameters
        return out

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

    @property
    def purelist_isregular(self):
        return True

    @property
    def purelist_depth(self):
        return len(self.inner_shape) + 1

    @property
    def is_identity_like(self):
        return False

    @property
    def minmax_depth(self):
        depth = len(self.inner_shape) + 1
        return (depth, depth)

    @property
    def branch_depth(self):
        return (False, len(self.inner_shape) + 1)

    @property
    def fields(self):
        return []

    @property
    def is_tuple(self):
        return False

    @property
    def dimension_optiontype(self):
        return False

    def _columns(self, path, output, list_indicator):
        output.append(".".join(path))

    def _select_columns(self, match_specifier):
        return self

    def _prune_columns(self, is_inside_record_or_union: bool) -> Self:
        return self

    def _column_types(self):
        return (ak.types.numpytype.primitive_to_dtype(self._primitive),)



from __future__ import annotations

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
# ruff: noqa: B023
__all__ = ("enforce_type",)
from itertools import permutations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._nplikes.shape import unknown_length
from awkward._parameters import type_parameters_equal
from awkward._typing import NamedTuple
from awkward.types.numpytype import primitive_to_dtype

np = NumpyMetadata.instance()


@high_level_function
def enforce_type(array, type, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        type (#ak.types.Type, or str): The type that `array` will be enforced to.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array whose structure is modified to match the given type.

    In addition to preserving the existing type and/or changing parameters,

    - #ak.types.OptionType can be added

      >>> a = ak.Array([1, 2, 3])
      >>> a.type.show()
      3 * int64
      >>> b = ak.enforce_type(a, "?int64")
      >>> b.type.show()
      3 * ?int64

      or removed (if there are no missing values)

      >>> a = ak.Array([1, 2, 3, None])
      >>> b = a[:-1]
      >>> b.type.show()
      3 * ?int64
      >>> c = ak.enforce_type(b, "int64")
      >>> c.type.show()
      3 * int64

    - #ak.types.UnionType can

      * grow to include new variant types,

      >>> a = ak.Array([{'x': 1}, 2.0])
      >>> a.type.show()
      2 * union[
          {
              x: int64
          },
          float64
      ]
      >>> b = ak.enforce_type(a, "union[{x: int64}, float64, string]")
      >>> b.type.show()
      2 * union[
      {
          x: float32
      },
          float64,
          string
      ]

      * convert to a single type,

      >>> a = ak.concatenate([
      ...   ak.Array([{'x': 1}, {'x': 2}]),
      ...   ak.Array([{'x': True, "y": None}, {'x': False, "y": None}])
      ... ])
      >>> a.type.show()
      4 * union[
          {
              x: int64
          },
          {
              x: bool,
              y: ?unknown
          }
      ]
      >>> b = ak.enforce_type(a, "{x: float64}")
      >>> b.type.show()
      4 * {
          x: float64
      }

      * project to a single type (if conversion to a single type is not possible, and the union contains no values for this type),

      >>> a = ak.concatenate([
      ...   ak.Array([{'x': 1}, {'x': 2}]),
      ...   ak.Array([{'x': "yes", "y": None}, {'x': "no", "y": None}])
      ... ])
      >>> b = a[:2]
      >>> b.type.show()
      2 * union[
          {
              x: int64
          },
          {
              x: string,
              y: ?unknown
          }
      ]
      >>> c = ak.enforce_type(b, "{x: int64}")
      >>> c.type.show()
      2 * {
          x: int64
      }

      * change type in a single variant.

      >>> a = ak.Array([{'x': 1}, 2.0])
      >>> a.type.show()
      2 * union[
          {
              x: int64
          },
          float64
      ]
      >>> b = ak.enforce_type(a, "union[{x: float32}, float64]")
      >>> b.type.show()
      2 * union[
      {
          x: float32
      },
          float64
      ]

      Due to these rules, changes to more than one variant of a union must be performed with multiple calls to #ak.enforce_type

    - #ak.types.RecordType can

      * grow to include new optional fields / slots,

      >>> a = ak.Array([{'x': 1}])
      >>> a.type.show()
      1 * {
          x: int64
      }
      >>> b = ak.enforce_type(a, "{x: int64, y: ?float32}")
      >>> b.type.show()
      1 * {
          x: int64,
          y: ?float32
      }

      * shrink to drop existing fields / slots.

      >>> a = ak.Array([{'x': 1, 'y': 1j+3}])
      >>> a.type.show()
      1 * {
          x: int64,
          y: complex128
      }
      >>> b = ak.enforce_type(a, "{x: int64}")
      >>> b.type.show()
      1 * {
          x: int64
      }

      A #ak.types.RecordType may only be converted to another #ak.types.RecordType if it is of the same flavour, i.e.
      tuples can be converted to tuples, or records to records. Where a new field/slot is added to a #ak.types.RecordType,
      it must be an #ak.types.OptionType. For tuples, slots may only be added to the end of the tuple
    - #ak.types.RegularType can convert to a #ak.types.ListType

      >>> a = ak.to_regular([[1, 2, 3], [4, 5, 6]])
      >>> a.type.show()
      2 * 3 * int64
      >>> b = ak.enforce_type(a, "var * int64")
      >>> b.type.show()
      2 * var * int64

    - #ak.types.ListType can convert to a #ak.types.RegularType

      >>> a = ak.Array([[1, 2, 3], [4, 5, 6]])
      >>> a.type.show()
      2 * var * int64
      >>> b = ak.enforce_type(a, "3 * int64")
      >>> b.type.show()
      2 * 3 * int64

    - #ak.types.NumpyType can change primitive

      >>> a = ak.Array([1, 2, 3])
      >>> a.type.show()
      3 * int64
      >>> b = ak.enforce_type(a, "float32")
      >>> b.type.show()
      3 * float32

    - #ak.types.UnknownType can be converted to any other type

      >>> a = ak.Array([])
      >>> a.type.show()
      0 * unknown
      >>> b = ak.enforce_type(a, "float32")
      >>> b.type.show()
      0 * float32

      and can be converted to from any other type.

      >>> a = ak.Array([1, 2, 3])
      >>> a.type.show()
      3 * int64
      >>> b = ak.enforce_type(a, "?unknown")
      >>> b.type.show()
      3 * ?unknown

    The conversion rules outlined above are not data-dependent; the appropriate rule is chosen from the layout and the
    given type value. If the conversion is not possible given the layout data, e.g. a conversion from an irregular list
    to a regular type, it will fail.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, type, highlevel, behavior)


def _impl(array, type_, highlevel, behavior):
    layout = ak.to_layout(array, allow_record=True)

    if isinstance(type_, str):
        type_ = ak.types.from_datashape(type_, highlevel=False)

    if isinstance(type_, (ak.types.ArrayType, ak.types.ScalarType)):
        raise TypeError(
            "High-level type objects are not supported by this function. Instead, "
            "a low-level type object (instance of ak.types.Type) should be used. "
            "If you are using a high-level type `type` from another array (e.g. using `array.type`), "
            "then the low-level type object can be found under `type.content`"
        )

    # Ensure we re-wrap records!
    if isinstance(layout, ak.record.Record):
        out = ak.record.Record(
            _enforce_type(layout.array[layout.at : layout.at + 1], type_), 0
        )
    else:
        out = _enforce_type(layout, type_)

    return wrap_layout(
        out,
        like=array,
        behavior=behavior,
        highlevel=highlevel,
    )


# TODO: move this if it ends up being useful elsewhere
def _layout_has_type(layout: ak.contents.Content, type_: ak.types.Type) -> bool:
    """
    Args:
        layout: content object
        type_: low-level type object

    Returns True if the layout satisfies the given type;, otherwise False.
    """
    if not type_parameters_equal(layout._parameters, type_._parameters):
        return False

    if layout.is_unknown:
        return isinstance(type_, ak.types.UnknownType)
    elif layout.is_option:
        return isinstance(type_, ak.types.OptionType) and _layout_has_type(
            layout.content, type_.content
        )
    elif layout.is_indexed:
        return _layout_has_type(layout.content, type_)
    elif layout.is_regular:
        return (
            isinstance(type_, ak.types.RegularType)
            and (
                layout.size is unknown_length
                or type_.size is unknown_length
                or layout.size == type_.size
            )
            and _layout_has_type(layout.content, type_.content)
        )
    elif layout.is_list:
        return isinstance(type_, ak.types.ListType) and _layout_has_type(
            layout.content, type_.content
        )
    elif layout.is_numpy:
        for _ in range(layout.purelist_depth - 1):
            if not isinstance(type_, ak.types.RegularType):
                return False
            type_ = type_.content
        return isinstance(
            type_, ak.types.NumpyType
        ) and layout.dtype == primitive_to_dtype(type_.primitive)
    elif layout.is_record:
        if (
            not isinstance(type_, ak.types.RecordType)
            or type_.is_tuple != layout.is_tuple
        ):
            return False

        if layout.is_tuple:
            return all(
                _layout_has_type(c, t) for c, t in zip(layout.contents, type_.contents)
            )
        else:
            return (frozenset(layout.fields) == frozenset(type_.fields)) and all(
                _layout_has_type(layout.content(f), type_.content(f))
                for f in type_.fields
            )
    elif layout.is_union:
        if len(layout.contents) != len(type_.contents):
            return False

        for contents in permutations(layout.contents):
            if all(
                _layout_has_type(layout, type_)
                for layout, type_ in zip(contents, type_.contents)
            ):
                return True
        return False
    else:
        raise TypeError(layout)


# TODO: move this if it ends up being useful elsewhere
class _TypeEnforceableResult(NamedTuple):
    is_enforceable: bool
    requires_packing: bool


def _type_is_enforceable(
    layout: ak.contents.Content, type_: ak.types.Type
) -> _TypeEnforceableResult:
    """
    Determine whether the layout can be enforced to the given type.

    Args:
        layout: content object
        type_: low-level type object

    Returns: a namedtuple with `is_enforceable` and `requires_packing` fields.
    """
    if layout.is_unknown:
        return _TypeEnforceableResult(is_enforceable=True, requires_packing=False)

    elif isinstance(type_, ak.types.UnknownType):
        return _TypeEnforceableResult(is_enforceable=False, requires_packing=False)

    elif layout.is_option:
        if isinstance(type_, ak.types.OptionType):
            # Converting to an unknown inside an option!
            if isinstance(type_.content, ak.types.UnknownType):
                return _TypeEnforceableResult(
                    is_enforceable=True, requires_packing=False
                )
            else:
                return _type_is_enforceable(layout.content, type_.content)
        else:
            content_is_enforceable, content_needs_packed = _type_is_enforceable(
                layout.content, type_
            )
            return _TypeEnforceableResult(
                is_enforceable=content_is_enforceable, requires_packing=True
            )

    elif layout.is_indexed:
        return _type_is_enforceable(layout.content, type_)

    # Here we *don't* have any layouts that are options, unknowns, or indexed
    # If we see an option, we are therefore *adding* one
    elif isinstance(type_, ak.types.OptionType):
        # Converting to an unknown inside an option!
        if isinstance(type_.content, ak.types.UnknownType):
            return _TypeEnforceableResult(is_enforceable=True, requires_packing=False)
        else:
            return _type_is_enforceable(layout, type_.content)

    elif layout.is_union:
        # If the target is a union type, then we have to determine the solution for e.g.
        # {A, B, C, D} → {X, Y, C, Z}.
        if isinstance(type_, ak.types.UnionType):
            n_type_contents = len(type_.contents)
            n_layout_contents = len(layout.contents)
            # The general operation of converting between one union and another can be decomposed into multiple
            # conversions from {A, B, C} to {A, B}, or from {A, B} to {A, B, C}.
            # Here we will only allow these transformations, because (a) the code is easier to write!
            # and (b) it is much easier to reason about _as a user_.

            # If the target has _more_ contents, then we assume we can add an unused content
            # Assume here that we have a *subset* of the type, i.e layout is {A, B, C}
            # and type is {A, B, C, D, ...}.
            if n_type_contents > n_layout_contents:
                # We can permute the type order, as union contents are not ordered
                # Permute the index, so that we can later recover remaining types
                ix_contents = range(n_type_contents)
                for ix_perm_contents in permutations(ix_contents, n_layout_contents):
                    retained_types = [type_.contents[j] for j in ix_perm_contents]
                    # Require that all layouts match types for layout permutation
                    if all(
                        _layout_has_type(c, t)
                        for c, t in zip(layout.contents, retained_types)
                    ):
                        return _TypeEnforceableResult(
                            is_enforceable=True, requires_packing=False
                        )

                return _TypeEnforceableResult(
                    is_enforceable=False, requires_packing=False
                )

            # Otherwise, we assume that we're projecting out one (or more) of our contents
            # Assume here that we have a *subset* of the layout, i.e layout is {A, B, C, D, ...}
            # and type is {A, B, C}. As the layout needs to lose a content, we must hope that the matching
            # permutation (by type) is also one that drops only unused contents from the union,
            # as layout operation must be typetracer-predictable
            elif n_layout_contents > n_type_contents:
                ix_contents = range(n_layout_contents)
                for ix_perm_contents in permutations(ix_contents, n_type_contents):
                    retained_contents = [layout.contents[j] for j in ix_perm_contents]
                    # Require that all layouts match types for layout permutation
                    if all(
                        _layout_has_type(c, t)
                        for c, t in zip(retained_contents, type_.contents)
                    ):
                        return _TypeEnforceableResult(
                            is_enforceable=True, requires_packing=True
                        )
                return _TypeEnforceableResult(
                    is_enforceable=False, requires_packing=False
                )

            # Type and layout have same number of contents. Up-to *one* content can differ
            else:
                ix_contents = range(n_type_contents)
                for ix_perm_contents in permutations(ix_contents):
                    permuted_types = [type_.contents[j] for j in ix_perm_contents]

                    # How many contents match types in this permutation?
                    content_matches_type = [
                        _layout_has_type(c, t)
                        for c, t in zip(layout.contents, permuted_types)
                    ]
                    n_matching = sum(content_matches_type, 0)

                    if n_matching == len(type_.contents):
                        return _TypeEnforceableResult(
                            is_enforceable=True, requires_packing=False
                        )
                    # Single content differs, we can convert by position
                    elif n_matching == len(type_.contents) - 1:
                        for tag, content_type, is_match in zip(
                            range(len(layout.contents)),
                            permuted_types,
                            content_matches_type,
                        ):
                            if not is_match:
                                # This content is being converted
                                return _type_is_enforceable(
                                    layout.contents[tag], content_type
                                )
                        raise AssertionError()
                    else:
                        return _TypeEnforceableResult(
                            is_enforceable=False, requires_packing=False
                        )
        # Otherwise, we are projecting out the union to a single type
        else:
            contents_enforceable = [
                _type_is_enforceable(content, type_)
                for tag, content in enumerate(layout.contents)
            ]
            if all(c.is_enforceable for c in contents_enforceable):
                # Conversion mode
                return _TypeEnforceableResult(
                    is_enforceable=True,
                    requires_packing=any(
                        c.requires_packing for c in contents_enforceable
                    ),
                )
            else:
                # Projection mode
                # Find the first content whose type equals the given type
                for tag, content in enumerate(layout.contents):  # noqa: B007
                    if _layout_has_type(content, type_):
                        return _TypeEnforceableResult(
                            is_enforceable=True, requires_packing=False
                        )
                return _TypeEnforceableResult(
                    is_enforceable=False, requires_packing=False
                )

    # Here we *don't* have any layouts that are options, unknowns, indexed, or unions
    # If we see a union, we are therefore *adding* one
    elif isinstance(type_, ak.types.UnionType):
        for _i, content_type in enumerate(type_.contents):
            if _layout_has_type(layout, content_type):
                return _type_is_enforceable(layout, content_type)
        return _TypeEnforceableResult(is_enforceable=False, requires_packing=False)

    elif layout.is_regular:
        if isinstance(type_, ak.types.RegularType):
            if layout.size == type_.size:
                return _TypeEnforceableResult(
                    is_enforceable=False, requires_packing=False
                )
            return _type_is_enforceable(layout.content, type_.content)

        elif isinstance(type_, ak.types.ListType):
            return _type_is_enforceable(layout.content, type_.content)

        else:
            return _TypeEnforceableResult(is_enforceable=False, requires_packing=False)

    elif layout.is_list:
        if isinstance(type_, ak.types.RegularType):
            return _type_is_enforceable(layout.content, type_.content)

        elif isinstance(type_, ak.types.ListType):
            return _type_is_enforceable(layout.content, type_.content)

        else:
            return _TypeEnforceableResult(is_enforceable=False, requires_packing=False)

    elif layout.is_numpy:
        for _ in range(layout.purelist_depth - 1):
            if not isinstance(type_, ak.types.RegularType):
                return _TypeEnforceableResult(
                    is_enforceable=False, requires_packing=False
                )
            type_ = type_.content
        if isinstance(type_, ak.types.NumpyType):
            return _TypeEnforceableResult(
                is_enforceable=True,
                requires_packing=primitive_to_dtype(type_.primitive) != layout.dtype,
            )
        else:
            return _TypeEnforceableResult(
                is_enforceable=False,
                requires_packing=False,
            )

    elif layout.is_record:
        if isinstance(type_, ak.types.RecordType):
            if type_.is_tuple and layout.is_tuple:
                # Recurse into shared contents
                type_contents = iter(type_.contents)
                contents_enforceable = [
                    _type_is_enforceable(c, t)
                    for c, t in zip(layout.contents, type_contents)
                ]
                # Anything left in `type_contents` are the types of new slots
                for next_type in type_contents:
                    if not isinstance(next_type, ak.types.OptionType):
                        return _TypeEnforceableResult(
                            is_enforceable=False, requires_packing=False
                        )

                return _TypeEnforceableResult(
                    is_enforceable=all(c.is_enforceable for c in contents_enforceable),
                    requires_packing=any(
                        c.requires_packing for c in contents_enforceable
                    ),
                )

            elif not (type_.is_tuple or layout.is_tuple):
                layout_fields = frozenset(layout._fields)

                # Compute existing and new fields
                # Use lists to preserve type order
                existing_fields = []
                new_fields = []
                for field in type_.fields:
                    if field in layout_fields:
                        existing_fields.append(field)
                    else:
                        new_fields.append(field)

                # Recurse into shared contents
                contents_enforceable = [
                    _type_is_enforceable(layout.content(f), type_.content(f))
                    for f in existing_fields
                ]
                for field in new_fields:
                    field_type = type_.content(field)
                    if not isinstance(field_type, ak.types.OptionType):
                        return _TypeEnforceableResult(
                            is_enforceable=False, requires_packing=False
                        )

                return _TypeEnforceableResult(
                    is_enforceable=all(c.is_enforceable for c in contents_enforceable),
                    requires_packing=any(
                        c.requires_packing for c in contents_enforceable
                    ),
                )

            else:
                return _TypeEnforceableResult(
                    is_enforceable=False, requires_packing=False
                )
        else:
            return _TypeEnforceableResult(is_enforceable=False, requires_packing=False)
    else:
        raise TypeError(layout)


def _recurse_indexed_any(
    layout: ak.contents.IndexedArray, type_: ak.types.Type
) -> ak.contents.Content:
    content_enforceable = _type_is_enforceable(layout.content, type_)

    if content_enforceable.requires_packing:
        # To ensure that we can project out options, we need to know
        # exactly what's visible to the user, so we'll project our contents
        return _enforce_type(layout.project(), type_)
    else:
        # If the types match, then we don't need to project, as only parameters
        # are changed (if at all)
        return layout.copy(content=_enforce_type(layout.content, type_))


def _recurse_unknown_any(
    layout: ak.contents.EmptyArray, type_: ak.types.Type
) -> ak.contents.Content:
    type_form = ak.forms.from_type(type_)
    return type_form.length_zero_array(highlevel=False).copy(
        parameters=type_._parameters
    )


def _recurse_any_unknown(layout: ak.contents.Content, type_: ak.types.UnknownType):
    raise TypeError(
        "cannot convert non-EmptyArray layouts to a bare UnknownType. "
        "To introduce an UnknownType, it must be wrapped in an OptionType"
    )


def _recurse_option_any(
    layout: ak.contents.IndexedOptionArray
    | ak.contents.BitMaskedArray
    | ak.contents.ByteMaskedArray
    | ak.contents.UnmaskedArray,
    type_: ak.types.Type,
) -> ak.contents.Content:
    # option → option (no change)
    if isinstance(type_, ak.types.OptionType):
        # Converting to an unknown inside an option!
        if isinstance(type_.content, ak.types.UnknownType):
            return ak.contents.IndexedOptionArray(
                ak.index.Index64(layout.backend.index_nplike.full(layout.length, -1)),
                ak.forms.from_type(type_.content).length_zero_array(
                    backend=layout.backend, highlevel=False
                ),
            )
        else:
            # Check that we can build the content
            content_enforceable = _type_is_enforceable(layout.content, type_.content)
            if content_enforceable.requires_packing:
                # If so, convert to packed so that any non-referenced content items are trimmed
                # This is required so that unused union items are seen to be safe to project out later
                # We don't use to_packed(), as it recurses
                index_nplike = layout.backend.index_nplike
                new_index = index_nplike.empty(layout.length, dtype=np.int64)

                is_none = layout.mask_as_bool(False)
                num_none = index_nplike.count_nonzero(is_none)

                new_index[is_none] = -1
                new_index[~is_none] = index_nplike.arange(
                    layout.length - num_none,
                    dtype=new_index.dtype,
                )
                return ak.contents.IndexedOptionArray(
                    ak.index.Index64(new_index, nplike=index_nplike),
                    _enforce_type(layout.project(), type_.content),
                    parameters=layout._parameters,
                )
            else:
                return layout.copy(
                    content=_enforce_type(layout.content, type_.content),
                    parameters=type_._parameters,
                )

    # drop option!
    else:
        # Check that we can build the content
        content_enforceable = _type_is_enforceable(layout.content, type_)

        if layout.backend.index_nplike.any(layout.mask_as_bool(False)):
            raise ValueError(
                "option types can only be removed if there are no missing values"
            )
        # If the option-type is not indexed, then at this point it's already packed
        elif not layout.is_indexed:
            return _enforce_type(layout.content[: layout.length], type_)
        # Otherwise project out if required
        elif content_enforceable.requires_packing:
            return _enforce_type(layout.project(), type_)
        # If we don't need to be packed, we can create a simple IndexedArray
        else:
            # Convert option to IndexedOptionArray and determine index of valid values
            assert layout.is_indexed
            return ak.contents.IndexedArray.simplified(
                index=ak.index.Index64(layout.index.data[layout.mask_as_bool(True)]),
                content=_enforce_type(layout.content, type_),
                parameters=layout._parameters,
            )


def _recurse_any_option(
    layout: ak.contents.Content, type_: ak.types.OptionType
) -> ak.contents.Content:
    # Converting to an unknown inside an option!
    if isinstance(type_.content, ak.types.UnknownType):
        return ak.contents.IndexedOptionArray(
            ak.index.Index64(layout.backend.index_nplike.full(layout.length, -1)),
            ak.forms.from_type(type_.content).length_zero_array(
                backend=layout.backend, highlevel=False
            ),
        )
    else:
        return ak.contents.UnmaskedArray(
            _enforce_type(layout, type_.content), parameters=type_._parameters
        )


def _recurse_union_any(
    layout: ak.contents.UnionArray, type_: ak.types.Type
) -> ak.contents.Content:
    # If the target is a union type, then we have to determine the solution for e.g.
    # {A, B, C, D} → {X, Y, C, Z}.
    if isinstance(type_, ak.types.UnionType):
        return _recurse_union_union(layout, type_)
    # Otherwise, we are projecting out the union to a single type
    else:
        return _recurse_union_non_union(layout, type_)


def _recurse_union_union(
    layout: ak.contents.UnionArray, type_: ak.types.UnionType
) -> ak.contents.Content:
    n_type_contents = len(type_.contents)
    n_layout_contents = len(layout.contents)
    # The general operation of converting between one union and another can be decomposed into multiple
    # conversions from {A, B, C} to {A, B}, or from {A, B} to {A, B, C}.
    # Here we will only allow these transformations, because (a) the code is easier to write!
    # and (b) it is much easier to reason about _as a user_.

    # If the target has _more_ contents, then we assume we can add an unused content
    # Assume here that we have a *subset* of the type, i.e layout is {A, B, C}
    # and type is {A, B, C, D, ...}.
    if n_type_contents > n_layout_contents:
        # We can permute the type order, as union contents are not ordered
        # Permute the index, so that we can later recover remaining types
        ix_contents = range(n_type_contents)
        for ix_perm_contents in permutations(ix_contents, n_layout_contents):
            retained_types = [type_.contents[j] for j in ix_perm_contents]
            # Require that all layouts match types for layout permutation
            if all(
                _layout_has_type(c, t) for c, t in zip(layout.contents, retained_types)
            ):
                break

        else:
            # No permutation succeeded
            raise NotImplementedError(
                "UnionArray(s) can currently only be converted into UnionArray(s) with a greater number contents if the "
                "layout contents are equal to some permutation of the type contents "
            )
        ix_perm_contents_set = frozenset(ix_perm_contents)
        missing_types = [
            type_.contents[j] for j in ix_contents if j not in ix_perm_contents_set
        ]

        # We only need to recurse here to enable parameter changes
        # Given that we _know_ all layouts match their types for the permutation,
        # we don't need to project these contents — they won't be operated upon (besides parameters)
        contents = [
            _enforce_type(b, c) for b, c in zip(layout.contents, retained_types)
        ]
        contents.extend(
            [
                ak.forms.from_type(t).length_zero_array(
                    highlevel=False, backend=layout.backend
                )
                for t in missing_types
            ]
        )
        return layout.copy(contents=contents, parameters=type_._parameters)

    # Otherwise, we assume that we're projecting out one (or more) of our contents
    # Assume here that we have a *subset* of the layout, i.e layout is {A, B, C, D, ...}
    # and type is {A, B, C}. As the layout needs to lose a content, we must hope that the matching
    # permutation (by type) is also one that drops only unused contents from the union,
    # as layout operation must be typetracer-predictable
    elif n_layout_contents > n_type_contents:
        ix_contents = range(n_layout_contents)
        for ix_perm_contents in permutations(ix_contents, n_type_contents):
            retained_contents = [layout.contents[j] for j in ix_perm_contents]
            # Require that all layouts match types for layout permutation
            if all(
                _layout_has_type(c, t)
                for c, t in zip(retained_contents, type_.contents)
            ):
                break
        else:
            raise NotImplementedError(
                "UnionArray(s) can currently only be converted into UnionArray(s) with a greater "
                "number of contents if the layout contents are compatible with some permutation of "
                "the type contents"
            )

        # We only need to recurse here to enable parameter changes
        # Given that we _know_ all layouts match their types for the permutation,
        # we don't need to project these contents — they won't be operated upon (besides parameters)
        contents = [
            _enforce_type(c, t) for c, t in zip(retained_contents, type_.contents)
        ]

        is_trivial_permutation = ix_perm_contents == range(n_type_contents)
        # Optimisation: if this is the trivial permutation, swe don't need to do any tag re-arranging
        if is_trivial_permutation:
            layout_tags = layout.tags
        else:
            layout_tags = ak.index.Index8.empty(
                layout.tags.length, layout.backend.index_nplike
            )

        # Ensure that the union references all of the tags of the permutation,
        # and re-order the tags if this is not the trivial permutation
        _total_used_tags = 0
        for i, j in zip(ix_perm_contents, range(n_type_contents)):
            layout_tag_is_i = layout.tags.data == i

            # Rewrite the tags if they need to be condensed (i.e., not if this is the trivial permutation)
            if not is_trivial_permutation:
                layout_tags.data[layout_tag_is_i] = j

            # Keep track of the length of layout subcontent
            _total_used_tags += layout.backend.index_nplike.count_nonzero(
                layout_tag_is_i
            )
        # Is the new union of the same length as the original?
        total_used_tags = layout.backend.index_nplike.index_as_shape_item(
            _total_used_tags
        )
        if not (
            total_used_tags is unknown_length
            or layout.length is unknown_length
            or total_used_tags == layout.length
        ):
            raise ValueError("union conversion must not be lossless")

        return layout.copy(
            tags=layout_tags,
            contents=contents,
            parameters=type_._parameters,
        )

    # Type and layout have same number of contents. Up-to *one* content can differ
    else:
        ix_contents = range(n_type_contents)
        for ix_perm_contents in permutations(ix_contents):
            permuted_types = [type_.contents[j] for j in ix_perm_contents]

            # How many contents match types in this permutation?
            content_matches_type = [
                _layout_has_type(c, t) for c, t in zip(layout.contents, permuted_types)
            ]
            n_matching = sum(content_matches_type, 0)

            # If all contents are nominally equal to the position-matched type, then only parameters have changed
            if n_matching == len(type_.contents):
                # Now build the result
                contents = [
                    _enforce_type(c, t) for c, t in zip(layout.contents, permuted_types)
                ]
                return layout.copy(
                    contents=contents,
                    parameters=type_._parameters,
                )
            # Single content differs, we can convert by position
            elif n_matching == len(type_.contents) - 1:
                next_contents = []
                index: ak.index.Index | None = None
                for tag, content_type, is_match in zip(
                    range(len(layout.contents)), permuted_types, content_matches_type
                ):
                    # If the types agree between the intended type and content, then include this content
                    # as-is, only recursing to update parameters. Because the types agree, we're safe
                    # not to project out this content
                    if is_match:
                        # assert not builder_needs_packed(builder)
                        next_contents.append(
                            _enforce_type(layout.contents[tag], content_type)
                        )
                    # Otherwise, this content is being converted, and we need to recurse into the projection
                    # to ensure that the content is packed
                    else:
                        # tag_needs_packed = builder_needs_packed(builder)
                        tag_content_enforceable = _type_is_enforceable(
                            layout.contents[tag], content_type
                        )

                        if tag_content_enforceable.requires_packing:
                            # Project changed ccontent
                            layout_content = layout.project(tag)
                            # Rebuild the index as an enumeration over the (dense) projection
                            # This ensures that it is packed!
                            index_data = layout.backend.index_nplike.asarray(
                                layout.index.data, copy=True
                            )
                            is_tag = layout.tags.data == tag
                            index_data[is_tag] = layout.backend.index_nplike.arange(
                                layout_content.length, dtype=index_data.dtype
                            )
                            index = ak.index.Index(index_data)
                            next_contents.append(
                                _enforce_type(layout_content, content_type)
                            )
                        else:
                            index = layout.index
                            next_contents.append(
                                _enforce_type(layout.contents[tag], content_type)
                            )

                return layout.copy(
                    index=index,
                    contents=next_contents,
                    parameters=type_._parameters,
                )
            else:
                raise TypeError(
                    "UnionArray(s) can currently only be converted into UnionArray(s) with the same number of contents "
                    "if no greater than one content differs in type"
                )


def _recurse_union_non_union(
    layout: ak.contents.UnionArray, type_: ak.types.Type
) -> ak.contents.Content:
    index_nplike = layout.backend.index_nplike
    if all(_type_is_enforceable(c, type_).is_enforceable for c in layout.contents):
        # Convert each projected content to the required type
        next_contents = []
        index_data = index_nplike.empty(layout.length, dtype=np.int64)
        j = 0
        for tag in range(len(layout.contents)):
            tag_content = layout.project(tag)
            # Set the index of these tags to a simple range
            i, j = j, j + index_nplike.shape_item_as_index(tag_content.length)
            index_data[layout.tags.data == tag] = index_nplike.arange(
                i, j, dtype=np.int64
            )
            # Convert layout
            next_contents.append(_enforce_type(tag_content, type_))

        # Merge the results
        for content in next_contents[1:]:
            assert ak._do.mergeable(next_contents[0], content, mergebool=False)
        next_content = ak._do.mergemany(next_contents)

        # Index over them
        index = ak.index.Index64(index_data)
        return ak.contents.IndexedArray(index, next_content)
    else:
        # Find the first content whose type equals the given type
        for tag, content in enumerate(layout.contents):  # noqa: B007
            if _layout_has_type(content, type_):
                break
        else:
            raise TypeError(
                f"UnionArray(s) can only be converted into {type_} if it has the same type, "
                f"but no content with type {type_} was found"
            )

        # Require that we are the only content
        content_is_tag = layout.tags.data == tag
        if index_nplike.known_data and not index_nplike.all(content_is_tag):
            raise ValueError(
                f"UnionArray(s) can only be converted to {type_} if they are equivalent to their "
                f"projections"
            )
        else:
            # We don't need to pack, as the type hasn't changed, so introduce an indexed type.
            # This ensures that unions over records don't needlessly project.
            # From the canonical rules, the content of a union *can* be an index, so we use simplified
            return ak.contents.IndexedArray.simplified(
                ak.index.Index(layout.index.data[content_is_tag]),
                _enforce_type(content, type_),
            )


def _recurse_any_union(
    layout: ak.contents.Content, type_: ak.types.UnionType
) -> ak.contents.Content:
    index_nplike = layout.backend.index_nplike

    for i, content_type in enumerate(type_.contents):
        if not _layout_has_type(layout, content_type):
            continue

        # Build zero-length contents from the new types
        tags = index_nplike.zeros(layout.length, dtype=np.int8)
        index = index_nplike.arange(layout.length, dtype=np.int64)

        other_contents = [
            ak.forms.from_type(t).length_zero_array(
                backend=layout.backend, highlevel=False
            )
            for j, t in enumerate(type_.contents)
            if j != i
        ]

        return ak.contents.UnionArray(
            tags=ak.index.Index8(tags, nplike=index_nplike),
            index=ak.index.Index64(index, nplike=index_nplike),
            contents=[
                _enforce_type(layout, content_type),
                *other_contents,
            ],
            parameters=type_._parameters,
        )

    raise TypeError(
        f"{type(layout).__name__} can only be converted into a UnionType if it is compatible with one "
        "of its contents, but no compatible content as found"
    )


def _recurse_regular_any(
    layout: ak.contents.RegularArray, type_: ak.types.Type
) -> ak.contents.Content:
    if isinstance(type_, ak.types.RegularType):
        # regular → regular requires same size!
        if layout.size != type_.size:
            raise ValueError(
                f"regular layout has different size ({layout.size}) to type ({type_.size})"
            )

        return layout.copy(
            content=_enforce_type(layout.content, type_.content),
            parameters=type_._parameters,
        )

    elif isinstance(type_, ak.types.ListType):
        layout_list = layout.to_ListOffsetArray64(True)
        return layout_list.copy(
            content=_enforce_type(layout.content, type_.content),
            parameters=type_._parameters,
        )

    else:
        raise TypeError(
            f"lists can only be converted to lists, options of lists, or unions thereof, not {type_}"
        )


def _recurse_list_any(
    layout: ak.contents.ListArray | ak.contents.ListOffsetArray, type_: ak.types.Type
) -> ak.contents.Content:
    if isinstance(type_, ak.types.RegularType):
        layout_regular = layout.to_RegularArray()

        # Empty arrays can have any size!
        if layout_regular.length is not unknown_length and layout_regular.length == 0:
            return layout_regular.copy(
                # Correct the size
                size=type_.size,
                content=_enforce_type(layout_regular.content, type_.content),
            )

        elif layout_regular.size is unknown_length or layout_regular.size == type_.size:
            return layout_regular.copy(
                # The result of `to_RegularArray` should already be packed
                content=_enforce_type(layout_regular.content, type_.content),
                parameters=type_._parameters,
            )

        else:
            # Runtime error
            raise ValueError(
                f"converted regular layout has different size ({layout_regular.size}) to type ({type_.size})"
            )

    elif isinstance(type_, ak.types.ListType):
        content_enforceable = _type_is_enforceable(layout.content, type_.content)
        if content_enforceable.requires_packing:
            # Need to pack the content!
            layout = layout.to_ListOffsetArray64(True)
            layout = layout[: layout.offsets[-1]]
            return layout.copy(
                content=_enforce_type(layout.content, type_.content),
                parameters=type_._parameters,
            )
        else:
            # Don't need to pack the content
            return layout.copy(
                content=_enforce_type(layout.content, type_.content),
                parameters=type_._parameters,
            )

    else:
        raise TypeError(
            f"lists can only be converted to lists, options of lists, or unions thereof, not {type_}"
        )


def _recurse_numpy_any(
    layout: ak.contents.NumpyArray, type_: ak.types.Type
) -> ak.contents.Content:
    if len(layout.shape) == 1:
        if not isinstance(type_, ak.types.NumpyType):
            raise TypeError(
                "NumpyArray(s) can only be converted into NumpyArray(s), options of NumpyArray(s), or "
                "unions thereof"
            )
        return ak.values_astype(
            # HACK: drop parameters from type so that character arrays are supported
            layout.copy(parameters=None),
            to=primitive_to_dtype(type_.primitive),
            highlevel=False,
        ).copy(parameters=type_._parameters)

    else:
        assert len(layout.shape) > 0
        return _enforce_type(layout.to_RegularArray(), type_)


def _recurse_record_any(
    layout: ak.contents.RecordArray, type_: ak.types.Type
) -> ak.contents.Content:
    if isinstance(type_, ak.types.RecordType):
        if type_.is_tuple and layout.is_tuple:
            # Recurse into shared contents
            type_contents = iter(type_.contents)
            next_contents = [
                _enforce_type(c, t) for c, t in zip(layout.contents, type_contents)
            ]
            # Anything left in `type_contents` are the types of new slots
            for next_type in type_contents:
                # Added types must be options, so that they can be all-None
                if not isinstance(next_type, ak.types.OptionType):
                    raise TypeError(
                        "can only add new slots to a tuple if they are option types"
                    )
                # Append new contents
                next_contents.append(
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64(
                            layout.backend.index_nplike.full(layout.length, -1)
                        ),
                        ak.forms.from_type(next_type.content).length_zero_array(
                            backend=layout.backend, highlevel=False
                        ),
                    )
                )

            return layout.copy(
                fields=None,
                contents=next_contents,
                parameters=type_._parameters,
            )

        elif not (type_.is_tuple or layout.is_tuple):
            layout_fields = frozenset(layout._fields)

            # Compute existing and new fields
            # Use lists to preserve type order
            existing_fields = []
            new_fields = []
            for field in type_.fields:
                if field in layout_fields:
                    existing_fields.append(field)
                else:
                    new_fields.append(field)
            next_fields = existing_fields + new_fields

            # Recurse into shared contents
            next_contents = [
                _enforce_type(layout.content(f), type_.content(f))
                for f in existing_fields
            ]
            for field in new_fields:
                field_type = type_.content(field)

                # Added types must be options, so that they can be all-None
                if not isinstance(field_type, ak.types.OptionType):
                    raise TypeError(
                        "can only add new fields to a record if they are option types"
                    )
                # Append new contents
                next_contents.append(
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64(
                            layout.backend.index_nplike.full(layout.length, -1)
                        ),
                        ak.forms.from_type(field_type.content).length_zero_array(
                            backend=layout.backend, highlevel=False
                        ),
                    )
                )

            return layout.copy(
                fields=next_fields,
                contents=next_contents,
                parameters=type_._parameters,
            )

        else:
            raise ValueError(
                "RecordArray(s) cannot be converted between records and tuples."
            )
    else:
        raise TypeError(
            "RecordArray(s) can only be converted into RecordArray(s), options of RecordArray(s), or "
            "unions thereof"
        )


def _enforce_type(
    layout: ak.contents.Content, type_: ak.types.Type
) -> ak.contents.Content:
    """
    Args:
        layout: layout to recurse into
        type_: expected type of the recursion result

    Returns a callable which converts from `layout` to `type_`. layout ensures that
    calling `recurse` is a type-only program, whilst keeping the conversion logic
    adjacent to the type check logic.
    """
    if layout.is_unknown:
        return _recurse_unknown_any(layout, type_)

    elif isinstance(type_, ak.types.UnknownType):
        return _recurse_any_unknown(layout, type_)

    elif layout.is_option:
        return _recurse_option_any(layout, type_)

    elif layout.is_indexed:
        return _recurse_indexed_any(layout, type_)

    # Here we *don't* have any layouts that are options, unknowns, or indexed
    # If we see an option, we are therefore *adding* one
    elif isinstance(type_, ak.types.OptionType):
        return _recurse_any_option(layout, type_)

    elif layout.is_union:
        return _recurse_union_any(layout, type_)

    # Here we *don't* have any layouts that are options, unknowns, indexed, or unions
    # If we see a union, we are therefore *adding* one
    elif isinstance(type_, ak.types.UnionType):
        return _recurse_any_union(layout, type_)

    elif layout.is_regular:
        return _recurse_regular_any(layout, type_)

    elif layout.is_list:
        return _recurse_list_any(layout, type_)

    elif layout.is_numpy:
        return _recurse_numpy_any(layout, type_)

    elif layout.is_record:
        return _recurse_record_any(layout, type_)
    else:
        raise NotImplementedError(type(layout), type_)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")

to_list = ak.operations.to_list


def test():
    import awkward._connect.pyarrow

    assert to_list(
        awkward._connect.pyarrow.handle_arrow(
            pyarrow.Table.from_pydict({"x": [None, None, None]})
        )
    ) == [{"x": None}, {"x": None}, {"x": None}]
    assert to_list(
        awkward._connect.pyarrow.handle_arrow(
            pyarrow.Table.from_pydict({"x": [[None, None], [], [None]]})
        )
    ) == [{"x": [None, None]}, {"x": []}, {"x": [None]}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import numpy.testing
import pytest

import awkward as ak

jax = pytest.importorskip("jax")
jax.config.update("jax_platform_name", "cpu")
jax.config.update("jax_enable_x64", True)

ak.jax.register_and_check()

# #### ak.contents.NumpyArray ####


test_regulararray = ak.Array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], backend="jax"
)
test_regulararray_tangent = ak.Array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], backend="jax"
)

test_regulararray_jax = jax.numpy.array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=np.float64
)
test_regulararray_tangent_jax = jax.numpy.array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=np.float64
)


@pytest.mark.parametrize("axis", [0, 1, None])
@pytest.mark.parametrize("func_ak", [ak.sum, ak.prod, ak.min, ak.max])
def test_reducer(func_ak, axis):
    func_jax = getattr(jax.numpy, func_ak.__name__)

    def func_ak_with_axis(x):
        return func_ak(x, axis=axis)

    def func_jax_with_axis(x):
        return func_jax(x, axis=axis)

    value_jvp, jvp_grad = jax.jvp(
        func_ak_with_axis, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_jax_with_axis, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_ak_with_axis, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_jax_with_axis, test_regulararray_jax)

    numpy.testing.assert_allclose(
        ak.to_list(value_jvp), value_jvp_jax.tolist(), rtol=1e-9, atol=np.inf
    )
    numpy.testing.assert_allclose(
        ak.to_list(value_vjp), value_vjp_jax.tolist(), rtol=1e-9, atol=np.inf
    )
    numpy.testing.assert_allclose(
        ak.to_list(jvp_grad), jvp_grad_jax.tolist(), rtol=1e-9, atol=np.inf
    )
    numpy.testing.assert_allclose(
        ak.to_list(vjp_func(value_vjp)[0]),
        (vjp_func_jax(value_vjp_jax)[0]).tolist(),
        rtol=1e-9,
        atol=np.inf,
    )


@pytest.mark.parametrize("axis", [None])
@pytest.mark.parametrize("func_ak", [ak.any, ak.all])
def test_bool_returns(func_ak, axis):
    func_jax = getattr(jax.numpy, func_ak.__name__)

    def func_ak_with_axis(x):
        return func_ak(x, axis=axis)

    def func_jax_with_axis(x):
        return func_jax(x, axis=axis)

    value_jvp, jvp_grad = jax.jvp(
        func_ak_with_axis, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_jax_with_axis, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_ak_with_axis, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_jax_with_axis, test_regulararray_jax)

    assert jvp_grad.dtype == jvp_grad_jax.dtype

    assert value_jvp.tolist() == value_jvp_jax.tolist()
    assert value_vjp.tolist() == value_vjp_jax.tolist()

    numpy.testing.assert_allclose(
        ak.to_list(vjp_func(value_vjp)[0]),
        (vjp_func_jax(value_vjp_jax)[0]).tolist(),
        rtol=1e-9,
        atol=np.inf,
    )


@pytest.mark.parametrize("axis", [0, 1, -1])
@pytest.mark.parametrize("func_ak", [ak.any, ak.all])
def test_bool_raises(func_ak, axis):
    def func_with_axis(x):
        return func_ak(x, axis=axis)

    with pytest.raises(
        TypeError, match=".*Make sure that you are not computing the derivative.*"
    ):
        jax.jvp(func_with_axis, (test_regulararray,), (test_regulararray_tangent,))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from awkward._behavior import find_array_typestr
from awkward._errors import deprecate
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._parameters import parameters_are_equal, type_parameters_equal
from awkward._regularize import is_integer
from awkward._typing import Self, final
from awkward._util import UNSET
from awkward.types.type import Type


@final
class RegularType(Type):
    def copy(
        self,
        *,
        content: Type = UNSET,
        size: ShapeItem = UNSET,
        parameters=UNSET,
        typestr=UNSET,
    ) -> Self:
        return RegularType(
            self._content if content is UNSET else content,
            size=self._size if size is UNSET else size,
            parameters=self._parameters if parameters is UNSET else parameters,
            typestr=self._typestr if typestr is UNSET else typestr,
        )

    def __init__(self, content, size, *, parameters=None, typestr=None):
        if not isinstance(content, Type):
            raise TypeError(
                "{} 'content' must be a Type subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not (size is unknown_length or (is_integer(size) and size >= 0)):
            raise ValueError(
                "{} 'size' must be a non-negative int or None, not {}".format(
                    type(self).__name__, repr(size)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{} 'parameters' must be of type dict or None, not {}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not isinstance(typestr, str):
            raise TypeError(
                "{} 'typestr' must be of type string or None, not {}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._content = content
        self._size = size
        self._parameters = parameters
        self._typestr = typestr

    @property
    def content(self):
        return self._content

    @property
    def size(self):
        return self._size

    def _get_typestr(self, behavior) -> str | None:
        typestr = find_array_typestr(behavior, self._parameters, self._typestr)
        if typestr is not None:
            return typestr

        if self._parameters is None:
            return None

        name = self._parameters.get("__array__")
        if name == "string":
            return "string"
        elif name == "bytestring":
            return "bytes"
        else:
            return None

    def _str(self, indent, compact, behavior):
        if self._typestr is not None:
            deprecate("typestr argument is deprecated", "2.4.0")

        typestr = self._get_typestr(behavior)
        if typestr is not None:
            out = [f"{typestr}[{self._size}]"]

        else:
            params = self._str_parameters()

            if params is None:
                out = [
                    str(self._size),
                    " * ",
                    *self._content._str(indent, compact, behavior),
                ]
            else:
                out = [
                    "[",
                    str(self._size),
                    " * ",
                    *self._content._str(indent, compact, behavior),
                    ", ",
                    params,
                    "]",
                ]

        return [self._str_categorical_begin(), *out, self._str_categorical_end()]

    def __repr__(self):
        args = [repr(self._content), repr(self._size), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _is_equal_to(self, other, all_parameters: bool) -> bool:
        compare_parameters = (
            parameters_are_equal if all_parameters else type_parameters_equal
        )
        return (
            isinstance(other, type(self))
            and compare_parameters(self._parameters, other._parameters)
            and (self._size == other._size)
            and self._content._is_equal_to(other._content, all_parameters)
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import itertools

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_numpyarray():
    for dtype1, dtype2, dtype3, dtype4 in itertools.product(
        ("i1", "i2", "i4", "i8", "u1", "u2", "u4", "u8", "f4", "f8", "?"), repeat=4
    ):
        one = np.array([0, 1, 2], dtype=dtype1)
        two = np.array([3, 0], dtype=dtype2)
        three = np.array([], dtype=dtype3)
        four = np.array([4, 5, 0, 6, 7], dtype=dtype4)
        combined = np.concatenate([one, two, three, four])

        ak_combined = ak.contents.NumpyArray(one)._mergemany(
            [
                ak.contents.NumpyArray(two),
                ak.contents.NumpyArray(three),
                ak.contents.NumpyArray(four),
            ]
        )

        assert to_list(ak_combined) == combined.tolist()
        assert ak_combined.dtype == combined.dtype

        assert (
            ak.contents.NumpyArray(one)
            .to_typetracer()
            ._mergemany(
                [
                    ak.contents.NumpyArray(two),
                    ak.contents.NumpyArray(three),
                    ak.contents.NumpyArray(four),
                ]
            )
            .form
            == ak.contents.NumpyArray(one)
            ._mergemany(
                [
                    ak.contents.NumpyArray(two),
                    ak.contents.NumpyArray(three),
                    ak.contents.NumpyArray(four),
                ]
            )
            .form
        )

        ak_combined = ak.contents.NumpyArray(one)._mergemany(
            [
                ak.contents.NumpyArray(two),
                ak.contents.EmptyArray(),
                ak.contents.NumpyArray(four),
            ]
        )

        assert to_list(ak_combined) == combined.tolist()
        assert ak_combined.dtype == np.concatenate([one, two, four]).dtype

        assert (
            ak.contents.NumpyArray(one)
            .to_typetracer()
            ._mergemany(
                [
                    ak.contents.NumpyArray(two),
                    ak.contents.EmptyArray(),
                    ak.contents.NumpyArray(four),
                ]
            )
            .form
            == ak.contents.NumpyArray(one)
            ._mergemany(
                [
                    ak.contents.NumpyArray(two),
                    ak.contents.EmptyArray(),
                    ak.contents.NumpyArray(four),
                ]
            )
            .form
        )


def test_lists():
    one = ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout
    two = ak.highlevel.Array([[1.1, 2.2], [3.3, 4.4]]).layout
    three = ak.contents.EmptyArray()
    four = ak.operations.from_numpy(
        np.array([[10], [20]]), regulararray=True, highlevel=False
    )
    assert to_list(one._mergemany([two, three, four])) == [
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [10.0],
        [20.0],
    ]
    assert to_list(four._mergemany([three, two, one])) == [
        [10.0],
        [20.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )
    assert (
        four.to_typetracer()._mergemany([three, four, one]).form
        == four._mergemany([three, four, one]).form
    )

    one = ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout
    two = ak.highlevel.Array([[1.1, 2.2], [3.3, 4.4]]).layout
    one = ak.contents.ListArray(one.starts, one.stops, one.content)
    two = ak.contents.ListArray(two.starts, two.stops, two.content)
    assert to_list(one._mergemany([two, three, four])) == [
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [10.0],
        [20.0],
    ]
    assert to_list(four._mergemany([three, two, one])) == [
        [10.0],
        [20.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
    ]

    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )

    assert (
        four.to_typetracer()._mergemany([three, four, one]).form
        == four._mergemany([three, four, one]).form
    )


def test_records():
    one = ak.highlevel.Array(
        [{"x": 1, "y": [1]}, {"x": 2, "y": [1, 2]}, {"x": 3, "y": [1, 2, 3]}]
    ).layout
    two = ak.highlevel.Array([{"y": [], "x": 4}, {"y": [3, 2, 1], "x": 5}]).layout
    three = two[0:0]
    four = ak.highlevel.Array([{"x": 6, "y": [1]}, {"x": 7, "y": [1, 2]}]).layout
    # ak.highlevel.Array([{"x": 6, "y": [1]}, {"x": 7, "y": [1, 2]}]).layout
    assert to_list(one._mergemany([two, three, four])) == [
        {"x": 1, "y": [1]},
        {"x": 2, "y": [1, 2]},
        {"x": 3, "y": [1, 2, 3]},
        {"y": [], "x": 4},
        {"y": [3, 2, 1], "x": 5},
        {"x": 6, "y": [1]},
        {"x": 7, "y": [1, 2]},
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )

    three = ak.contents.EmptyArray()
    assert to_list(one._mergemany([two, three, four])) == [
        {"x": 1, "y": [1]},
        {"x": 2, "y": [1, 2]},
        {"x": 3, "y": [1, 2, 3]},
        {"y": [], "x": 4},
        {"y": [3, 2, 1], "x": 5},
        {"x": 6, "y": [1]},
        {"x": 7, "y": [1, 2]},
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )


def test_tuples():
    one = ak.highlevel.Array([(1, [1]), (2, [1, 2]), (3, [1, 2, 3])]).layout
    two = ak.highlevel.Array([(4, []), (5, [3, 2, 1])]).layout
    three = two[0:0]
    four = ak.highlevel.Array([(6, [1]), (7, [1, 2])]).layout
    assert to_list(one._mergemany([two, three, four])) == [
        (1, [1]),
        (2, [1, 2]),
        (3, [1, 2, 3]),
        (4, []),
        (5, [3, 2, 1]),
        (6, [1]),
        (7, [1, 2]),
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )

    three = ak.contents.EmptyArray()
    assert to_list(one._mergemany([two, three, four])) == [
        (1, [1]),
        (2, [1, 2]),
        (3, [1, 2, 3]),
        (4, []),
        (5, [3, 2, 1]),
        (6, [1]),
        (7, [1, 2]),
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )


def test_indexed():
    one = ak.highlevel.Array([1, 2, 3, None, 4, None, None, 5]).layout
    two = ak.highlevel.Array([6, 7, 8]).layout
    three = ak.contents.EmptyArray()
    four = ak.highlevel.Array([9, None, None]).layout
    assert to_list(one._mergemany([two, three, four])) == [
        1,
        2,
        3,
        None,
        4,
        None,
        None,
        5,
        6,
        7,
        8,
        9,
        None,
        None,
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )


def test_reverse_indexed():
    one = ak.highlevel.Array([1, 2, 3]).layout
    two = ak.highlevel.Array([4, 5]).layout
    three = ak.highlevel.Array([None, 6, None]).layout
    assert to_list(one._mergemany([two, three])) == [1, 2, 3, 4, 5, None, 6, None]
    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )

    four = ak.highlevel.Array([7, 8, None, None, 9]).layout
    assert to_list(one._mergemany([two, three, four])) == [
        1,
        2,
        3,
        4,
        5,
        None,
        6,
        None,
        7,
        8,
        None,
        None,
        9,
    ]
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )


def test_bytemasked():
    one = ak.contents.ByteMaskedArray(
        ak.index.Index8([True, True, False, True, False, True]),
        ak.highlevel.Array([1, 2, 3, 4, 5, 6]).layout,
        valid_when=True,
    )
    two = ak.contents.ByteMaskedArray(
        ak.index.Index8([True, False, False, True, True]),
        ak.highlevel.Array([7, 99, 999, 8, 9]).layout,
        valid_when=True,
    )
    three = ak.highlevel.Array([100, 200, 300]).layout
    four = ak.highlevel.Array([None, None, 123, None]).layout
    assert to_list(one._mergemany([two, three, four])) == [
        1,
        2,
        None,
        4,
        None,
        6,
        7,
        None,
        None,
        8,
        9,
        100,
        200,
        300,
        None,
        None,
        123,
        None,
    ]
    assert to_list(four._mergemany([three, two, one])) == [
        None,
        None,
        123,
        None,
        100,
        200,
        300,
        7,
        None,
        None,
        8,
        9,
        1,
        2,
        None,
        4,
        None,
        6,
    ]
    assert to_list(three._mergemany([four, one])) == [
        100,
        200,
        300,
        None,
        None,
        123,
        None,
        1,
        2,
        None,
        4,
        None,
        6,
    ]
    assert to_list(three._mergemany([four, one, two])) == [
        100,
        200,
        300,
        None,
        None,
        123,
        None,
        1,
        2,
        None,
        4,
        None,
        6,
        7,
        None,
        None,
        8,
        9,
    ]
    assert to_list(three._mergemany([two, one])) == [
        100,
        200,
        300,
        7,
        None,
        None,
        8,
        9,
        1,
        2,
        None,
        4,
        None,
        6,
    ]
    assert to_list(three._mergemany([two, one, four])) == [
        100,
        200,
        300,
        7,
        None,
        None,
        8,
        9,
        1,
        2,
        None,
        4,
        None,
        6,
        None,
        None,
        123,
        None,
    ]

    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )
    assert (
        four.to_typetracer()._mergemany([three, two, one]).form
        == four._mergemany([three, two, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([four, one]).form
        == three._mergemany([four, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([four, one, two]).form
        == three._mergemany([four, one, two]).form
    )
    assert (
        three.to_typetracer()._mergemany([two, one]).form
        == three._mergemany([two, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([two, one, four]).form
        == three._mergemany([two, one, four]).form
    )


def test_empty():
    one = ak.contents.EmptyArray()
    two = ak.contents.EmptyArray()
    three = ak.highlevel.Array([1, 2, 3]).layout
    four = ak.highlevel.Array([4, 5]).layout
    assert to_list(one._mergemany([two])) == []
    assert to_list(one._mergemany([two, one, two, one, two])) == []
    assert to_list(one._mergemany([two, three])) == [1, 2, 3]
    assert to_list(one._mergemany([two, three, four])) == [1, 2, 3, 4, 5]
    assert to_list(one._mergemany([three])) == [1, 2, 3]
    assert to_list(one._mergemany([three, four])) == [1, 2, 3, 4, 5]
    assert to_list(one._mergemany([three, two])) == [1, 2, 3]
    assert to_list(one._mergemany([three, two, four])) == [1, 2, 3, 4, 5]
    assert to_list(one._mergemany([three, four, two])) == [1, 2, 3, 4, 5]

    assert (
        one.to_typetracer()._mergemany([two.to_typetracer()]).form
        == one._mergemany([two]).form
    )
    assert (
        one.to_typetracer()._mergemany([two, one, two, one, two]).form
        == one._mergemany([two, one, two, one, two]).form
    )
    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )
    assert (
        one.to_typetracer()._mergemany([two, three, four]).form
        == one._mergemany([two, three, four]).form
    )
    assert (
        one.to_typetracer()._mergemany([three.to_typetracer()]).form
        == one._mergemany([three]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, four]).form
        == one._mergemany([three, four]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, two]).form
        == one._mergemany([three, two]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, two, four]).form
        == one._mergemany([three, two, four]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, four, two]).form
        == one._mergemany([three, four, two]).form
    )


def test_union():
    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, 200, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, 6, 7]).layout

    assert to_list(one._mergemany([two, three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(one._mergemany([three, two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        200,
        300,
    ]
    assert to_list(two._mergemany([one, three])) == [
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(two._mergemany([three, one])) == [
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(three._mergemany([one, two])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
    ]
    assert to_list(three._mergemany([two, one])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, two]).form
        == one._mergemany([three, two]).form
    )
    assert (
        two.to_typetracer()._mergemany([one, three]).form
        == two._mergemany([one, three]).form
    )
    assert (
        two.to_typetracer()._mergemany([three, one]).form
        == two._mergemany([three, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([one, two]).form
        == three._mergemany([one, two]).form
    )
    assert (
        three.to_typetracer()._mergemany([two, one]).form
        == three._mergemany([two, one]).form
    )


def test_union_option():
    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, None, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, 6, 7]).layout

    assert to_list(one._mergemany([two, three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(one._mergemany([three, two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        None,
        300,
    ]
    assert to_list(two._mergemany([one, three])) == [
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(two._mergemany([three, one])) == [
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(three._mergemany([one, two])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
    ]
    assert to_list(three._mergemany([two, one])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, two]).form
        == one._mergemany([three, two]).form
    )
    assert (
        two.to_typetracer()._mergemany([one, three]).form
        == two._mergemany([one, three]).form
    )
    assert (
        two.to_typetracer()._mergemany([three, one]).form
        == two._mergemany([three, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([one, two]).form
        == three._mergemany([one, two]).form
    )
    assert (
        three.to_typetracer()._mergemany([two, one]).form
        == three._mergemany([two, one]).form
    )

    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, None, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, None, 7]).layout

    assert to_list(one._mergemany([two, three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(one._mergemany([three, two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        None,
        300,
    ]
    assert to_list(two._mergemany([one, three])) == [
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(two._mergemany([three, one])) == [
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(three._mergemany([one, two])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
    ]
    assert to_list(three._mergemany([two, one])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, two]).form
        == one._mergemany([three, two]).form
    )
    assert (
        two.to_typetracer()._mergemany([one, three]).form
        == two._mergemany([one, three]).form
    )
    assert (
        two.to_typetracer()._mergemany([three, one]).form
        == two._mergemany([three, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([one, two]).form
        == three._mergemany([one, two]).form
    )
    assert (
        three.to_typetracer()._mergemany([two, one]).form
        == three._mergemany([two, one]).form
    )

    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, 200, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, None, 7]).layout

    assert to_list(one._mergemany([two, three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(one._mergemany([three, two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        200,
        300,
    ]
    assert to_list(two._mergemany([one, three])) == [
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(two._mergemany([three, one])) == [
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(three._mergemany([one, two])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
    ]
    assert to_list(three._mergemany([two, one])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )
    assert (
        one.to_typetracer()._mergemany([three, two]).form
        == one._mergemany([three, two]).form
    )
    assert (
        two.to_typetracer()._mergemany([one, three]).form
        == two._mergemany([one, three]).form
    )
    assert (
        two.to_typetracer()._mergemany([three, one]).form
        == two._mergemany([three, one]).form
    )
    assert (
        three.to_typetracer()._mergemany([one, two]).form
        == three._mergemany([one, two]).form
    )
    assert (
        three.to_typetracer()._mergemany([two, one]).form
        == three._mergemany([two, one]).form
    )


def test_strings():
    one = ak.highlevel.Array(["uno", "dos", "tres"]).layout
    two = ak.highlevel.Array(["un", "deux", "trois", "quatre"]).layout
    three = ak.highlevel.Array(["onay", "ootay", "eethray"]).layout
    assert to_list(one._mergemany([two, three])) == [
        "uno",
        "dos",
        "tres",
        "un",
        "deux",
        "trois",
        "quatre",
        "onay",
        "ootay",
        "eethray",
    ]

    assert (
        one.to_typetracer()._mergemany([two, three]).form
        == one._mergemany([two, three]).form
    )


def test_concatenate():
    one = ak.highlevel.Array([1, 2, 3]).layout
    two = ak.highlevel.Array([4.4, 5.5]).layout
    three = ak.highlevel.Array([6, 7, 8]).layout
    four = ak.highlevel.Array([[9, 9, 9], [10, 10, 10]]).layout

    assert to_list(ak.operations.concatenate([one, two, three, four])) == [
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
        [9, 9, 9],
        [10, 10, 10],
    ]
    assert to_list(ak.operations.concatenate([four, one, two, three])) == [
        [9, 9, 9],
        [10, 10, 10],
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
    ]
    assert to_list(ak.operations.concatenate([one, two, four, three])) == [
        1,
        2,
        3,
        4.4,
        5.5,
        [9, 9, 9],
        [10, 10, 10],
        6,
        7,
        8,
    ]

    five = ak.highlevel.Array(["nine", "ten"]).layout
    assert to_list(ak.operations.concatenate([one, two, three, five])) == [
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
        "nine",
        "ten",
    ]
    assert to_list(ak.operations.concatenate([five, one, two, three])) == [
        "nine",
        "ten",
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
    ]
    assert to_list(ak.operations.concatenate([one, two, five, three])) == [
        1,
        2,
        3,
        4.4,
        5.5,
        "nine",
        "ten",
        6,
        7,
        8,
    ]
